#cython: boundscheck=False
#cython: wraparound=False
#cython: language_level=3

import numpy as np
cimport numpy as np
import cython

ctypedef np.complex128_t complex_t
ctypedef np.float64_t real_t
ctypedef np.int64_t int_t

ctypedef fused T:
    real_t
    complex_t

cdef extern from "complex.h" nogil:
    double complex exp(double complex)

cdef extern from "math.h" nogil:
    double sin(double)
    double cos(double)

def evaluate_2D(b, u, list P, int r2c, int M, int start):
    if P[0].dtype == complex and P[1].dtype == complex:
        # 2 Fourier spaces
        if r2c < 0: # No R2C
            b = _evaluate_2D_cc0(b, u, P[0], P[1], r2c, M, start)
        else:
            b = _evaluate_2D_cc1(b, u, P[0], P[1], r2c, M, start)

    elif P[0].dtype == float and P[1].dtype == complex:
        # One non-fourier in axis=0
        if r2c < 0: # No R2C
            b = _evaluate_2D_rc0(b, u, P[0], P[1], r2c, M, start)
        else:
            b = _evaluate_2D_rc1(b, u, P[0], P[1], r2c, M, start)

    elif P[0].dtype == complex and P[1].dtype == float:
        # One non-fourier in axis=1
        if r2c < 0: # No R2C
            b = _evaluate_2D_cr0(b, u, P[0], P[1], r2c, M, start)
        else:
            b = _evaluate_2D_cr1(b, u, P[0], P[1], r2c, M, start)

    elif P[0].dtype == float and P[1].dtype == float:
        # Two non-fourier bases
        b = _evaluate_2D_rr0(b, u, P[0], P[1])

    return b

def _evaluate_2D_cc0(np.ndarray[complex_t, ndim=1] b,
                     np.ndarray[complex_t, ndim=2] u,
                     np.ndarray[complex_t, ndim=2] P0,
                     np.ndarray[complex_t, ndim=2] P1,
                     int r2c, int M, int start):
    cdef int k, l, i, kk, ll
    cdef complex_t p

    assert r2c < 0

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                b[i] = b[i] + u[k, l] * P0[i, k] * P1[i, l]
    return b

def _evaluate_2D_cc1(np.ndarray[real_t, ndim=1] b,
                     np.ndarray[complex_t, ndim=2] u,
                     np.ndarray[complex_t, ndim=2] P0,
                     np.ndarray[complex_t, ndim=2] P1,
                     int r2c, int M, int start):
    cdef int k, l, i, ii
    cdef real_t p
    assert r2c == 0 or r2c == 1
    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                p = (u[k, l] * P0[i, k] * P1[i, l]).real
                b[i] += p
                if r2c == 0:
                    ii = k + start
                else:
                    ii = l + start
                if ii > 0 & ii < M:
                    b[i] += p

    return b

def _evaluate_2D_rc0(np.ndarray[complex_t, ndim=1] b,
                     np.ndarray[complex_t, ndim=2] u,
                     np.ndarray[real_t, ndim=2] P0,
                     np.ndarray[complex_t, ndim=2] P1,
                     int r2c, int M, int start):
    cdef int k, l, i

    assert r2c < 0
    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                b[i] = b[i] + u[k, l] * P0[i, k] * P1[i, l]
    return b


def _evaluate_2D_rc1(np.ndarray[real_t, ndim=1] b,
                     np.ndarray[complex_t, ndim=2] u,
                     np.ndarray[real_t, ndim=2] P0,
                     np.ndarray[complex_t, ndim=2] P1,
                     int r2c, int M, int start):
    cdef int k, l, i, ii
    cdef real_t p
    assert r2c == 0 or r2c == 1

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                p = (u[k, l] * P0[i, k] * P1[i, l]).real
                b[i] += p
                if r2c == 0:
                    ii = k + start
                else:
                    ii = l + start
                if ii > 0 & ii < M:
                    b[i] += p
    return b

def _evaluate_2D_cr0(np.ndarray[complex_t, ndim=1] b,
                     np.ndarray[complex_t, ndim=2] u,
                     np.ndarray[complex_t, ndim=2] P0,
                     np.ndarray[real_t, ndim=2] P1,
                     int r2c, int M, int start):
    cdef int k, l, i

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                b[i] += u[k, l] * P0[i, k] * P1[i, l]

    return b

def _evaluate_2D_cr1(np.ndarray[real_t, ndim=1] b,
                     np.ndarray[complex_t, ndim=2] u,
                     np.ndarray[complex_t, ndim=2] P0,
                     np.ndarray[real_t, ndim=2] P1,
                     int r2c, int M, int start):
    cdef int k, l, i, ii
    cdef real_t p
    assert r2c == 0 or r2c == 1

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                p = (u[k, l] * P0[i, k] * P1[i, l]).real
                b[i] += p
                if r2c == 0:
                    ii = k + start
                else:
                    ii = l + start
                if ii > 0 & ii < M:
                    b[i] += p

    return b

def _evaluate_2D_rr0(np.ndarray[real_t, ndim=1] b,
                     np.ndarray[real_t, ndim=2] u,
                     np.ndarray[real_t, ndim=2] P0,
                     np.ndarray[real_t, ndim=2] P1):
    cdef int k, l, i

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                b[i] = b[i] + u[k, l] * P0[i, k] * P1[i, l]
    return b

def evaluate_3D(b, u, list P, int r2c, int M, int start):

    if P[0].dtype == complex and P[1].dtype == complex and P[2].dtype == complex:
        # Only Fourier bases
        if r2c < 0: # No R2C
            b = _evaluate_3D_ccc0(b, u, P[0], P[1], P[2], r2c, M, start)
        else:
            b = _evaluate_3D_ccc1(b, u, P[0], P[1], P[2], r2c, M, start)

    elif P[0].dtype == float and P[1].dtype == complex and P[2].dtype == complex:
        # One non-fourier in axis=0
        if r2c < 0: # No R2C
            b = _evaluate_3D_rcc0(b, u, P[0], P[1], P[2], r2c, M, start)
        else:
            b = _evaluate_3D_rcc1(b, u, P[0], P[1], P[2], r2c, M, start)

    elif P[0].dtype == complex and P[1].dtype == float and P[2].dtype == complex:
        # One non-fourier in axis=1
        if r2c < 0: # No R2C
            b = _evaluate_3D_crc0(b, u, P[0], P[1], P[2], r2c, M, start)
        else:
            b = _evaluate_3D_crc1(b, u, P[0], P[1], P[2], r2c, M, start)

    elif P[0].dtype == complex and P[1].dtype == complex and P[2].dtype == float:
        # One non-fourier in axis=2
        if r2c < 0: # No R2C
            b = _evaluate_3D_ccr0(b, u, P[0], P[1], P[2], r2c, M, start)
        else:
            b = _evaluate_3D_ccr1(b, u, P[0], P[1], P[2], r2c, M, start)

    elif P[0].dtype == float and P[1].dtype == float and P[2].dtype == complex:
        b = _evaluate_3D_rrc1(b, u, P[0], P[1], P[2], r2c, M, start)

    elif P[0].dtype == float and P[1].dtype == complex and P[2].dtype == float:
        b = _evaluate_3D_rcr1(b, u, P[0], P[1], P[2], r2c, M, start)

    elif P[0].dtype == complex and P[1].dtype == float and P[2].dtype == float:
        b = _evaluate_3D_crr1(b, u, P[0], P[1], P[2], r2c, M, start)

    elif P[0].dtype == float and P[1].dtype == float and P[2].dtype == float:
        b = _evaluate_3D_rrr(b, u, P[0], P[1], P[2])

    return b

def _evaluate_3D_ccc0(np.ndarray[complex_t, ndim=1] b,
                      np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[complex_t, ndim=2] P0,
                      np.ndarray[complex_t, ndim=2] P1,
                      np.ndarray[complex_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i
    cdef np.ndarray[complex_t, ndim=3] c = np.zeros((b.shape[0], u.shape[0], u.shape[1]), dtype=complex)
    cdef np.ndarray[complex_t, ndim=2] c2 = np.zeros((b.shape[0], u.shape[0]), dtype=complex)

    #for k in range(u.shape[0]):
    #    for l in range(u.shape[1]):
    #        for m in range(u.shape[2]):
    #            for i in range(b.shape[0]):
    #                b[i] = b[i] + u[k, l, m] * P0[i, k] * P1[i, l] * P2[i, m]
    #return b
    for i in range(b.shape[0]):
        for k in range(u.shape[0]):
            for l in range(u.shape[1]):
                for m in range(u.shape[2]):
                    c[i, k, l] += u[k, l, m] * P2[i, m]

    for i in range(b.shape[0]):
        for k in range(u.shape[0]):
            for l in range(u.shape[1]):
                c2[i, k] += c[i, k, l] * P1[i, l]

    for i in range(b.shape[0]):
        for k in range(u.shape[0]):
            b[i] += c2[i, k] * P0[i, k]
    return b

def _evaluate_3D_ccc1(np.ndarray[real_t, ndim=1] b,
                      np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[complex_t, ndim=2] P0,
                      np.ndarray[complex_t, ndim=2] P1,
                      np.ndarray[complex_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef complex_t p1, p2
    cdef real_t p, ur, ui

    for i in range(b.shape[0]):
        for k in range(u.shape[0]):
            if r2c == 0:
                ii = k + start
            for l in range(u.shape[1]):
                p1 = P0[i, k] * P1[i, l]
                if r2c == 1:
                    ii = l + start
                for m in range(u.shape[2]):
                    #p = (u[k, l, m] * p1 * P2[i, m]).real
                    p2 = p1 * P2[i, m]
                    ur = u[k, l, m].real
                    ui = u[k, l, m].imag
                    p = ur*p2.real - ui*p2.imag
                    b[i] += p
                    if r2c == 2:
                        ii = m + start
                    if ii > 0 & ii < M:
                        b[i] += p

    return b

def _evaluate_3D_rcc0(np.ndarray[complex_t, ndim=1] b,
                      np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[real_t, ndim=2] P0,
                      np.ndarray[complex_t, ndim=2] P1,
                      np.ndarray[complex_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i
    cdef np.ndarray[complex_t, ndim=3] c = np.zeros((b.shape[0], u.shape[0], u.shape[1]), dtype=complex)
    cdef np.ndarray[complex_t, ndim=2] c2 = np.zeros((b.shape[0], u.shape[0]), dtype=complex)


    assert r2c < 0

    #for k in range(u.shape[0]):
    #    for l in range(u.shape[1]):
    #        for m in range(u.shape[2]):
    #            for i in range(b.shape[0]):
    #                b[i] += P0[i, k] * (u[k, l, m] * P1[i, l] * P2[i, m])

    for i in range(b.shape[0]):
        for k in range(u.shape[0]):
            for l in range(u.shape[1]):
                for m in range(u.shape[2]):
                    c[i, k, l] += u[k, l, m] * P2[i, m]

    for i in range(b.shape[0]):
        for k in range(u.shape[0]):
            for l in range(u.shape[1]):
                c2[i, k] += c[i, k, l] * P1[i, l]

    for i in range(b.shape[0]):
        for k in range(u.shape[0]):
            b[i] += c2[i, k] * P0[i, k]

    return b

def _evaluate_3D_rcc1(np.ndarray[real_t, ndim=1] b,
                      np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[real_t, ndim=2] P0,
                      np.ndarray[complex_t, ndim=2] P1,
                      np.ndarray[complex_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef real_t p
    cdef complex_t p0

    assert r2c == 1 or r2c == 2

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    #p = (u[k, l, m] * P0[i, k] * P1[i, l] * P2[i, m]).real
                    p0 = P1[i, l] * P2[i, m]
                    p = P0[i, k]*(u[k, l, m].real*p0.real - u[k, l, m].imag*p0.imag)
                    b[i] += p
                    if r2c == 1:
                        ii = l + start
                    else:
                        ii = m + start
                    if ii > 0 & ii < M:
                        b[i] += p
    return b

def _evaluate_3D_crc0(np.ndarray[complex_t, ndim=1] b,
                      np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[complex_t, ndim=2] P0,
                      np.ndarray[real_t, ndim=2] P1,
                      np.ndarray[complex_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i

    assert r2c < 0

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    b[i] = b[i] + u[k, l, m] * P0[i, k] * P1[i, l] * P2[i, m]

    return b

def _evaluate_3D_crc1(np.ndarray[real_t, ndim=1] b,
                      np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[complex_t, ndim=2] P0,
                      np.ndarray[real_t, ndim=2] P1,
                      np.ndarray[complex_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef real_t p
    cdef complex_t p0

    assert r2c == 0 or r2c == 2

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    #p = (u[k, l, m] * P0[i, k] * P1[i, l] * P2[i, m]).real
                    p0 = P0[i, k] * P2[i, m]
                    p = P1[i, l]*(u[k, l, m].real*p0.real - u[k, l, m].imag*p0.imag)
                    b[i] += p
                    if r2c == 0:
                        ii = k + start
                    else:
                        ii = m + start
                    if ii > 0 & ii < M:
                        b[i] += p
    return b

def _evaluate_3D_ccr0(np.ndarray[complex_t, ndim=1] b, np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[complex_t, ndim=2] P0,
                      np.ndarray[complex_t, ndim=2] P1,
                      np.ndarray[real_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i

    assert r2c < 0

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    b[i] = b[i] + u[k, l, m] * P0[i, k] * P1[i, l] * P2[i, m]
    return b

def _evaluate_3D_ccr1(np.ndarray[real_t, ndim=1] b,
                      np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[complex_t, ndim=2] P0,
                      np.ndarray[complex_t, ndim=2] P1,
                      np.ndarray[real_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef real_t p
    cdef complex_t p0

    assert r2c == 0 or r2c == 1

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    #p = (u[k, l, m] * P0[i, k] * P1[i, l] * P2[i, m]).real
                    p0 = P0[i, k] * P1[i, l]
                    p = P2[i, m]*(u[k, l, m].real*p0.real - u[k, l, m].imag*p0.imag)
                    b[i] += p
                    if r2c == 0:
                        ii = k + start
                    else:
                        ii = l + start
                    if ii > 0 & ii < M:
                        b[i] += p
    return b

def _evaluate_3D_rrc1(np.ndarray[real_t, ndim=1] b,
                      np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[real_t, ndim=2] P0,
                      np.ndarray[real_t, ndim=2] P1,
                      np.ndarray[complex_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef real_t p
    cdef complex_t p0

    assert r2c == 2

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    p = (u[k, l, m] * P0[i, k] * P1[i, l] * P2[i, m]).real
                    #p0 = P2[i, m]
                    #p = P0[i, k]*P1[i, l]*(u[k, l, m].real*p0.real - u[k, l, m].imag*p0.imag)
                    b[i] += p
                    ii = m + start
                    if ii > 0 & ii < M:
                        b[i] += p
    return b

def _evaluate_3D_rcr1(np.ndarray[real_t, ndim=1] b,
                      np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[real_t, ndim=2] P0,
                      np.ndarray[complex_t, ndim=2] P1,
                      np.ndarray[real_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef real_t p
    cdef complex_t p0

    assert r2c == 1

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    p = (u[k, l, m] * P0[i, k] * P1[i, l] * P2[i, m]).real
                    #p0 = P2[i, m]
                    #p = P0[i, k]*P1[i, l]*(u[k, l, m].real*p0.real - u[k, l, m].imag*p0.imag)
                    b[i] += p
                    ii = l + start
                    if ii > 0 & ii < M:
                        b[i] += p
    return b

def _evaluate_3D_crr1(np.ndarray[real_t, ndim=1] b,
                      np.ndarray[complex_t, ndim=3] u,
                      np.ndarray[complex_t, ndim=2] P0,
                      np.ndarray[real_t, ndim=2] P1,
                      np.ndarray[real_t, ndim=2] P2,
                      int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef real_t p
    cdef complex_t p0

    assert r2c == 0

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    p = (u[k, l, m] * P0[i, k] * P1[i, l] * P2[i, m]).real
                    #p0 = P2[i, m]
                    #p = P0[i, k]*P1[i, l]*(u[k, l, m].real*p0.real - u[k, l, m].imag*p0.imag)
                    b[i] += p
                    ii = k + start
                    if ii > 0 & ii < M:
                        b[i] += p
    return b

def _evaluate_3D_rrr(np.ndarray[real_t, ndim=1] b,
                      np.ndarray[real_t, ndim=3] u,
                      np.ndarray[real_t, ndim=2] P0,
                      np.ndarray[real_t, ndim=2] P1,
                      np.ndarray[real_t, ndim=2] P2):
    cdef int k, l, m, i, ii
    cdef real_t p

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    b[i] += u[k, l, m] * P0[i, k] * P1[i, l] * P2[i, m]

    return b

def evaluate_lm_2D(list bases, b, u, x0, x1, w0, w1, int r2c, int M, int start):

    if np.all([base.family() == 'fourier' for base in bases]):
        # 2 Fourier spaces
        if r2c < 0: # No R2C
            b = _evaluate_lm_2D_cc0(b, u, x0, x1, w0, w1, r2c, M, start)
        else:
            b = _evaluate_lm_2D_cc1(b, u, x0, x1, w0, w1, r2c, M, start)

    elif bases[1].family() == 'fourier':
        # One non-fourier in axis=0
        if r2c < 0: # No R2C
            b = _evaluate_lm_2D_rc0(b, u, x0, x1, w0, w1, bases, r2c, M, start)
        else:
            b = _evaluate_lm_2D_rc1(b, u, x0, x1, w0, w1, bases, r2c, M, start)

    elif bases[0].family() == 'fourier':
        # One non-fourier in axis=1
        if r2c < 0: # No R2C
            b = _evaluate_lm_2D_cr0(b, u, x0, x1, w0, w1, bases, r2c, M, start)
        else:
            b = _evaluate_lm_2D_cr1(b, u, x0, x1, w0, w1, bases, r2c, M, start)

    return b

def _evaluate_lm_2D_cc0(np.ndarray[complex_t, ndim=1] b,
                        np.ndarray[complex_t, ndim=2] u,
                        np.ndarray[real_t, ndim=1] x0,
                        np.ndarray[real_t, ndim=1] x1,
                        np.ndarray[real_t, ndim=1] w0,
                        np.ndarray[real_t, ndim=1] w1,
                        int r2c, int M, int start):
    cdef int k, l, i
    assert r2c < 0

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                #b[i] += u[k, l] * exp(1j*(w0[k]*x0[i] + w1[l]*x1[i]))
                xx = w0[k]*x0[i] + w1[l]*x1[i]
                yc = cos(xx)
                ys = sin(xx)
                ur = u[k, l].real
                ui = u[k, l].imag
                b[i].real = b[i].real + (ur * yc - ui * ys)
                b[i].imag = b[i].imag + (ur * ys + ui * yc)

    return b

def _evaluate_lm_2D_cc1(np.ndarray[real_t, ndim=1] b,
                        np.ndarray[complex_t, ndim=2] u,
                        np.ndarray[real_t, ndim=1] x0,
                        np.ndarray[real_t, ndim=1] x1,
                        np.ndarray[real_t, ndim=1] w0,
                        np.ndarray[real_t, ndim=1] w1,
                        int r2c, int M, int start):
    cdef int k, l, i, ii
    cdef real_t p, xx, yc, ys, ur, ui

    for k in range(u.shape[0]):
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                #p = (u[k, l] * exp(1j*(w0[k]*x0[i] + w1[l]*x1[i]))).real
                xx = w0[k]*x0[i] + w1[l]*x1[i]
                yc = cos(xx)
                ys = sin(xx)
                ur = u[k, l].real
                ui = u[k, l].imag
                p = (ur * yc - ui * ys)
                b[i] += p
                if r2c == 0:
                    ii = k + start
                elif r2c == 1:
                    ii = l + start
                if ii > 0 & ii < M:
                    b[i] += p
    return b

def _evaluate_lm_2D_rc0(np.ndarray[complex_t, ndim=1] b,
                        np.ndarray[complex_t, ndim=2] u,
                        np.ndarray[real_t, ndim=1] x0,
                        np.ndarray[real_t, ndim=1] x1,
                        np.ndarray[real_t, ndim=1] w0,
                        np.ndarray[real_t, ndim=1] w1,
                        list bases,
                        int r2c, int M, int start):
    cdef int k, l, i
    cdef real_t xx, yc, ys, ur, ui
    cdef np.ndarray[real_t, ndim=1] P0 = np.zeros(x0.shape[0], dtype=float)
    assert r2c < 0
    for k in range(u.shape[0]):
        P0 = bases[0].evaluate_basis(x0, k, P0)
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                #b[i] += u[k, l] * P0[i] * exp(1j*w1[l]*x1[i])
                xx = w1[l]*x1[i]
                yc = cos(xx)
                ys = sin(xx)
                ur = u[k, l].real
                ui = u[k, l].imag
                b[i].real = b[i].real + P0[i]*(ur * yc - ui * ys)
                b[i].imag = b[i].imag + P0[i]*(ur * ys + ui * yc)

    return b


def _evaluate_lm_2D_rc1(np.ndarray[real_t, ndim=1] b,
                        np.ndarray[complex_t, ndim=2] u,
                        np.ndarray[real_t, ndim=1] x0,
                        np.ndarray[real_t, ndim=1] x1,
                        np.ndarray[real_t, ndim=1] w0,
                        np.ndarray[real_t, ndim=1] w1,
                        list bases,
                        int r2c, int M, int start):
    cdef int k, l, i, ii
    cdef real_t p, xx, yc, ys, ur, ui
    cdef np.ndarray[real_t, ndim=1] P0 = np.zeros(x0.shape[0], dtype=float)
    for k in range(u.shape[0]):
        P0 = bases[0].evaluate_basis(x0, k, P0)
        for l in range(u.shape[1]):
            for i in range(b.shape[0]):
                #p = (u[k, l] * P0[i] * exp(1j*w1[l]*x1[i])).real
                xx = w1[l]*x1[i]
                yc = cos(xx)
                ys = sin(xx)
                ur = u[k, l].real
                ui = u[k, l].imag
                p = P0[i]*(ur * yc - ui * ys)
                b[i] += p
                if r2c == 0:
                    ii = k + start
                elif r2c == 1:
                    ii = l + start
                if ii > 0 & ii < M:
                    b[i] += p
    return b

def _evaluate_lm_2D_cr0(np.ndarray[complex_t, ndim=1] b,
                        np.ndarray[complex_t, ndim=2] u,
                        np.ndarray[real_t, ndim=1] x0,
                        np.ndarray[real_t, ndim=1] x1,
                        np.ndarray[real_t, ndim=1] w0,
                        np.ndarray[real_t, ndim=1] w1,
                        list bases,
                        int r2c, int M, int start):
    cdef int k, l, i, ii
    cdef real_t xx, yc, ys, ur, ui
    cdef np.ndarray[real_t, ndim=1] P0 = np.zeros(x0.shape[0], dtype=float)

    for l in range(u.shape[1]):
        P0 = bases[1].evaluate_basis(x1, l, P0)
        for k in range(u.shape[0]):
            for i in range(b.shape[0]):
                #b[i] += u[k, l] * P0[i] * exp(1j*w0[k]*x0[i])
                xx = w0[k]*x0[i]
                yc = cos(xx)
                ys = sin(xx)
                ur = u[k, l].real
                ui = u[k, l].imag
                b[i].real = b[i].real + P0[i]*(ur * yc - ui * ys)
                b[i].imag = b[i].imag + P0[i]*(ur * ys + ui * yc)

    return b

def _evaluate_lm_2D_cr1(np.ndarray[real_t, ndim=1] b,
                        np.ndarray[complex_t, ndim=2] u,
                        np.ndarray[real_t, ndim=1] x0,
                        np.ndarray[real_t, ndim=1] x1,
                        np.ndarray[real_t, ndim=1] w0,
                        np.ndarray[real_t, ndim=1] w1,
                        list bases,
                        int r2c, int M, int start):
    cdef int k, l, i, ii
    cdef real_t p, y2, yc, ys, ur, ui
    cdef np.ndarray[real_t, ndim=1] P0 = np.zeros(x0.shape[0], dtype=float)

    for l in range(u.shape[1]):
        P0 = bases[1].evaluate_basis(x1, l, P0)
        for k in range(u.shape[0]):
            for i in range(b.shape[0]):
                #p = (u[k, l] * P0[i] * exp(1j*w0[k]*x0[i])).real
                y2 = w0[k]*x0[i]
                yc = cos(y2)
                ys = sin(y2)
                ur = u[k, l].real
                ui = u[k, l].imag
                p = P0[i]*(ur * yc - ui * ys)
                b[i] += p
                if r2c == 0:
                    ii = k + start
                elif r2c == 1:
                    ii = l + start
                if ii > 0 & ii < M:
                    b[i] += p
    return b

def evaluate_lm_3D(list bases, b, u, x0, x1, x2, w0, w1, w2, int r2c, int M, int start):

    if np.all([base.family() == 'fourier' for base in bases]):
        # Only Fourier bases
        if r2c < 0: # No R2C
            b = _evaluate_lm_3D_ccc0(b, u, x0, x1, x2, w0, w1, w2, r2c, M, start)
        else:
            b = _evaluate_lm_3D_ccc1(b, u, x0, x1, x2, w0, w1, w2, r2c, M, start)

    elif not bases[0].family() == 'fourier':
        # One non-fourier in axis=0
        if r2c < 0: # No R2C
            b = _evaluate_lm_3D_rcc0(b, u, x0, x1, x2, w0, w1, w2, bases, r2c, M, start)
        else:
            b = _evaluate_lm_3D_rcc1(b, u, x0, x1, x2, w0, w1, w2, bases, r2c, M, start)

    elif not bases[1].family() == 'fourier':
        # One non-fourier in axis=1
        if r2c < 0: # No R2C
            b = _evaluate_lm_3D_crc0(b, u, x0, x1, x2, w0, w1, w2, bases, r2c, M, start)
        else:
            b = _evaluate_lm_3D_crc1(b, u, x0, x1, x2, w0, w1, w2, bases, r2c, M, start)

    elif not bases[2].family() == 'fourier':
        # One non-fourier in axis=2
        if r2c < 0: # No R2C
            b = _evaluate_lm_3D_ccr0(b, u, x0, x1, x2, w0, w1, w2, bases, r2c, M, start)
        else:
            b = _evaluate_lm_3D_ccr1(b, u, x0, x1, x2, w0, w1, w2, bases, r2c, M, start)

    return b

def _evaluate_lm_3D_ccc0(complex_t[::1] b,
                         complex_t[:, :, ::1] u,
                         real_t[::1] x0,
                         real_t[::1] x1,
                         real_t[::1] x2,
                         real_t[::1] w0,
                         real_t[::1] w1,
                         real_t[::1] w2,
                         int r2c, int M, int start):
    cdef int k, l, m, i
    cdef double xx, y0, y1, yc, ys, br, bi, ur, ui
    for i in range(b.shape[0]):
        br = 0.0
        bi = 0.0
        for k in range(u.shape[0]):
            y0 = w0[k]*x0[i]
            for l in range(u.shape[1]):
                y1 = y0 + w1[l]*x1[i]
                for m in range(u.shape[2]):
                    xx = y1 + w2[m]*x2[i]
                    #b[i] += u[k, l, m] * (cos(xx) + 1j*sin(xx))
                    yc = cos(xx)
                    ys = sin(xx)
                    ur = u[k, l, m].real
                    ui = u[k, l, m].imag
                    br += (ur * yc - ui * ys)
                    bi += (ur * ys + ui * yc)
        b[i] = br + bi*1j

    return b

def _evaluate_lm_3D_ccc1(np.ndarray[real_t, ndim=1] b,
                         np.ndarray[complex_t, ndim=3] u,
                         np.ndarray[real_t, ndim=1] x0,
                         np.ndarray[real_t, ndim=1] x1,
                         np.ndarray[real_t, ndim=1] x2,
                         np.ndarray[real_t, ndim=1] w0,
                         np.ndarray[real_t, ndim=1] w1,
                         np.ndarray[real_t, ndim=1] w2,
                         int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef real_t p, y0, y1, y2, yc, ys, ur, ui
    for i in range(b.shape[0]):
        for k in range(u.shape[0]):
            y0 = w0[k]*x0[i]
            for l in range(u.shape[1]):
                y1 = y0 + w1[l]*x1[i]
                for m in range(u.shape[2]):
                    #p = (u[k, l, m] * exp(1j*(y1+w2[m]*x2[i]))).real
                    y2 = y1+w2[m]*x2[i]
                    yc = cos(y2)
                    ys = sin(y2)
                    ur = u[k, l, m].real
                    ui = u[k, l, m].imag
                    p = (ur * yc - ui * ys)
                    b[i] += p
                    if r2c == 0:
                        ii = k + start
                    elif r2c == 1:
                        ii = l + start
                    elif r2c == 2:
                        ii = m + start
                    if ii > 0 & ii < M:
                        b[i] += p
    return b

def _evaluate_lm_3D_rcc0(np.ndarray[complex_t, ndim=1] b,
                         np.ndarray[complex_t, ndim=3] u,
                         np.ndarray[real_t, ndim=1] x0,
                         np.ndarray[real_t, ndim=1] x1,
                         np.ndarray[real_t, ndim=1] x2,
                         np.ndarray[real_t, ndim=1] w0,
                         np.ndarray[real_t, ndim=1] w1,
                         np.ndarray[real_t, ndim=1] w2,
                         list bases,
                         int r2c, int M, int start):
    cdef int k, l, m, i, ll, mm
    cdef real_t y2, yc, ys, ur, ui
    cdef np.ndarray[real_t, ndim=1] P0 = np.zeros(x0.shape[0], dtype=float)
    assert r2c == 1 or r2c == 2 or r2c < 0

    for k in range(u.shape[0]):
        P0 = bases[0].evaluate_basis(x0, k, P0)
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    #b[i] += u[k, l, m] * P0[i] * exp(1j*(w1[l]*x1[i] + w2[m]*x2[i]))
                    y2 = w1[l]*x1[i] + w2[m]*x2[i]
                    yc = cos(y2)
                    ys = sin(y2)
                    ur = u[k, l, m].real
                    ui = u[k, l, m].imag
                    b[i].real = b[i].real + P0[i]*(ur * yc - ui * ys)
                    b[i].imag = b[i].imag + P0[i]*(ur * ys + ui * yc)

    return b

def _evaluate_lm_3D_rcc1(np.ndarray[real_t, ndim=1] b,
                         np.ndarray[complex_t, ndim=3] u,
                         np.ndarray[real_t, ndim=1] x0,
                         np.ndarray[real_t, ndim=1] x1,
                         np.ndarray[real_t, ndim=1] x2,
                         np.ndarray[real_t, ndim=1] w0,
                         np.ndarray[real_t, ndim=1] w1,
                         np.ndarray[real_t, ndim=1] w2,
                         list bases,
                         int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef real_t p, y0, y1, y2, yc, ys, ur, ui
    cdef np.ndarray[real_t, ndim=1] P0 = np.zeros(x0.shape[0], dtype=float)
    assert r2c == 1 or r2c == 2

    for k in range(u.shape[0]):
        P0 = bases[0].evaluate_basis(x0, k, P0)
        for l in range(u.shape[1]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    #p = (u[k, l, m] * P0[i] * exp(1j*(w1[l]*x1[i] + w2[m]*x2[i]))).real
                    y2 = w1[l]*x1[i] + w2[m]*x2[i]
                    yc = cos(y2)
                    ys = sin(y2)
                    ur = u[k, l, m].real
                    ui = u[k, l, m].imag
                    p = P0[i]*(ur * yc - ui * ys)
                    b[i] += p
                    if r2c == 1:
                        ii = l + start
                    elif r2c == 2:
                        ii = m + start
                    if ii > 0 & ii < M:
                        b[i] += p
    return b

def _evaluate_lm_3D_crc0(np.ndarray[complex_t, ndim=1] b,
                         np.ndarray[complex_t, ndim=3] u,
                         np.ndarray[real_t, ndim=1] x0,
                         np.ndarray[real_t, ndim=1] x1,
                         np.ndarray[real_t, ndim=1] x2,
                         np.ndarray[real_t, ndim=1] w0,
                         np.ndarray[real_t, ndim=1] w1,
                         np.ndarray[real_t, ndim=1] w2,
                         list bases,
                         int r2c, int M, int start):
    cdef int k, l, m, i
    cdef real_t y2, yc, ys, ur, ui
    cdef np.ndarray[real_t, ndim=1] P0 = np.zeros(x0.shape[0], dtype=float)
    assert r2c < 0

    for l in range(u.shape[1]):
        P0 = bases[1].evaluate_basis(x1, l, P0)
        for k in range(u.shape[0]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    #b[i] += u[k, l, m] * P0[i] * exp(1j*(w0[k]*x0[i] + w2[m]*x2[i]))
                    y2 = w0[k]*x0[i] + w2[m]*x2[i]
                    yc = cos(y2)
                    ys = sin(y2)
                    ur = u[k, l, m].real
                    ui = u[k, l, m].imag
                    b[i].real = b[i].real + P0[i]*(ur * yc - ui * ys)
                    b[i].imag = b[i].imag + P0[i]*(ur * ys + ui * yc)
    return b

def _evaluate_lm_3D_crc1(np.ndarray[real_t, ndim=1] b,
                         np.ndarray[complex_t, ndim=3] u,
                         np.ndarray[real_t, ndim=1] x0,
                         np.ndarray[real_t, ndim=1] x1,
                         np.ndarray[real_t, ndim=1] x2,
                         np.ndarray[real_t, ndim=1] w0,
                         np.ndarray[real_t, ndim=1] w1,
                         np.ndarray[real_t, ndim=1] w2,
                         list bases,
                         int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef real_t p, y2, yc, ys, ur, ui
    cdef np.ndarray[real_t, ndim=1] P0 = np.zeros(x0.shape[0], dtype=float)
    assert r2c == 0 or r2c == 2

    for l in range(u.shape[1]):
        P0 = bases[1].evaluate_basis(x1, l, P0)
        for k in range(u.shape[0]):
            for m in range(u.shape[2]):
                for i in range(b.shape[0]):
                    #p = (u[k, l, m] * P0[i] * exp(1j*(w0[k]*x0[i] + w1[m]*x2[i]))).real
                    y2 = w0[k]*x0[i] + w1[m]*x2[i]
                    yc = cos(y2)
                    ys = sin(y2)
                    ur = u[k, l, m].real
                    ui = u[k, l, m].imag
                    p = P0[i]*(ur * yc - ui * ys)
                    b[i] += p
                    if r2c == 0:
                        ii = k + start
                    elif r2c == 2:
                        ii = m + start
                    if ii > 0 & ii < M:
                        b[i] += p
    return b

def _evaluate_lm_3D_ccr0(np.ndarray[complex_t, ndim=1] b,
                         np.ndarray[complex_t, ndim=3] u,
                         np.ndarray[real_t, ndim=1] x0,
                         np.ndarray[real_t, ndim=1] x1,
                         np.ndarray[real_t, ndim=1] x2,
                         np.ndarray[real_t, ndim=1] w0,
                         np.ndarray[real_t, ndim=1] w1,
                         np.ndarray[real_t, ndim=1] w2,
                         list bases,
                         int r2c, int M, int start):
    cdef int k, l, m, i
    cdef real_t y2, yc, ys, ur, ui
    cdef np.ndarray[real_t, ndim=1] P0 = np.zeros(x0.shape[0], dtype=float)
    assert r2c < 0

    for m in range(u.shape[2]):
        P0 = bases[2].evaluate_basis(x2, m, P0)
        for k in range(u.shape[0]):
            for l in range(u.shape[1]):
                for i in range(b.shape[0]):
                    #b[i] += u[k, l, m] * P0[i] * exp(1j*(w0[k]*x0[i] + w1[l]*x1[i]))
                    y2 = w0[k]*x0[i] + w1[l]*x1[i]
                    yc = cos(y2)
                    ys = sin(y2)
                    ur = u[k, l, m].real
                    ui = u[k, l, m].imag
                    b[i].real = b[i].real + P0[i]*(ur * yc - ui * ys)
                    b[i].imag = b[i].imag + P0[i]*(ur * ys + ui * yc)

    return b

def _evaluate_lm_3D_ccr1(np.ndarray[real_t, ndim=1] b,
                         np.ndarray[complex_t, ndim=3] u,
                         np.ndarray[real_t, ndim=1] x0,
                         np.ndarray[real_t, ndim=1] x1,
                         np.ndarray[real_t, ndim=1] x2,
                         np.ndarray[real_t, ndim=1] w0,
                         np.ndarray[real_t, ndim=1] w1,
                         np.ndarray[real_t, ndim=1] w2,
                         list bases,
                         int r2c, int M, int start):
    cdef int k, l, m, i, ii
    cdef real_t p, y2, yc, ys, ur, ui
    cdef np.ndarray[real_t, ndim=1] P0 = np.zeros(x0.shape[0], dtype=float)
    assert r2c == 0 or r2c == 1

    for m in range(u.shape[2]):
        P0 = bases[2].evaluate_basis(x2, m, P0)
        for k in range(u.shape[0]):
            for l in range(u.shape[1]):
                for i in range(b.shape[0]):
                    #p = (u[k, l, m] * P0[i] * exp(1j*(w0[k]*x0[i] + w1[l]*x1[i]))).real
                    y2 = w0[k]*x0[i] + w1[l]*x1[i]
                    yc = cos(y2)
                    ys = sin(y2)
                    ur = u[k, l, m].real
                    ui = u[k, l, m].imag
                    p = P0[i]*(ur * yc - ui * ys)
                    b[i] += p
                    if r2c == 0:
                        ii = k + start
                    elif r2c == 1:
                        ii = l + start
                    if ii > 0 & ii < M:
                        b[i] += p
    return b



#cython: boundscheck=False
#cython: language_level=3
cimport numpy as np
import numpy as np
cimport cython
import cython
from libc.stdlib cimport malloc, free
from copy import copy
from numpy.polynomial import chebyshev as n_cheb
from .transforms import FMMdirect1, FMMcheb, FMMdirect2, FMMdirect3, \
    FMMdirect4, _leg2cheb, _cheb2leg, _Lambda
from .transforms cimport get_number_of_blocks

np.import_array()

cdef class FMMLevel:
    cdef public int N
    cdef public int Nn
    cdef public int diags
    cdef public int L
    cdef public int s
    cdef public int maxs
    cdef public int use_direct
    cdef public np.ndarray D
    cdef public np.ndarray Mmin
    cdef public np.ndarray fk
    cdef public np.ndarray Nk
    cdef public np.ndarray T
    cdef public np.ndarray Th
    cdef public np.ndarray ThT

    def __init__(self, int N, int diagonals=8, domains=None, levels=None,
                 int l2c=1, int maxs=100, int use_direct=-1):
        cdef:
            int i, s, rest, h, Nd, level
            np.ndarray[long, ndim=1] doms

        from shenfun.legendre.dlt import getChebyshev, conversionmatrix
        self.N = N
        self.diags = diagonals
        self.maxs = maxs
        self.use_direct = use_direct
        if self.N <= self.use_direct:
            self.Nn = N
            return

        if domains is not None:
            if isinstance(domains, int):
                if levels is None:
                    doms = np.cumsum(domains**np.arange(16))
                    levels = np.where((N//2-diagonals)/doms < maxs)[0][0]
                self.D = np.full(levels, domains, dtype=int)

            else:
                domains = np.atleast_1d(domains)
                levels = len(domains)
                if domains[levels-1] == 1:
                    D0 = domains.copy()
                    for i in range(1000):
                        Nd = get_number_of_blocks(levels, <long*>np.PyArray_DATA(domains), len(domains))
                        s = (N//2-diagonals)//Nd
                        if s < maxs:
                            break
                        else:
                            domains = D0*(i+2)
                self.D = domains
        else:
            if levels is None:
                domains = max(2, int(np.log10(N)))
                doms = np.cumsum(domains**np.arange(16))
                levels = np.where((N//2-diagonals)/doms < maxs)[0][0]
                levels = max(1, levels)
            else:
                for i in range(2, 1000):
                    Nd = np.cumsum(i**np.arange(levels+1))[levels]
                    s = (N//2-diagonals)//Nd
                    if s < maxs:
                        domains = i
                        break
            self.D = np.full(levels, domains, dtype=int)
        self.L = max(1, levels)
        Nd = get_number_of_blocks(self.L, <long*>np.PyArray_DATA(self.D), len(self.D))
        # Create new N with minimal size according to diags and N
        # Pad input array with zeros
        s, rest = divmod(N//2-diagonals, Nd)
        Nn = N
        if rest > 0 or Nn%2 == 1:
            s += 1
            Nn = 2*(diagonals+s*Nd)

        self.Nn = Nn
        self.s = s
        fk = []
        Nk = []
        self.Mmin = np.zeros(self.L, dtype=int)
        for level in range(self.L-1, -1, -1):
            Nx = getChebyshev(level, self.D, s, diagonals, fk, Nk,  l2c)
            self.Mmin[level] = Nx
        self.fk = np.hstack(fk)
        self.Nk = np.array(Nk, dtype=int)
        Mmax = self.Mmin.max()
        TT = {}
        for level in range(self.L-1, -1, -1):
            if self.D[level] not in TT:
                TT[self.D[level]] = conversionmatrix(self.D[level], Mmax)
        self.Th = np.concatenate([TT[d] for d in np.unique(self.D)])
        self.ThT = np.array([copy(self.Th[i].transpose()) for i in range(self.Th.shape[0])])
        self.T = np.zeros((2, s, self.Mmin[-1]))
        Ti = n_cheb.chebvander(np.linspace(-1, 1, 2*s+1)[:-1], self.Mmin[-1]-1)
        self.T[0] = Ti[::2]
        self.T[1] = Ti[1::2]

cdef class Leg2Cheb(FMMLevel):
    cdef public np.ndarray _output_array
    cdef np.ndarray _a
    cdef int axis
    cdef int Mmax
    cdef tuple si

    def __cinit__(self, np.ndarray input_array, output_array=None, int axis=0,
                  int diagonals=8, domains=None, levels=None, int maxs=100, int use_direct=-1):
        self._output_array = input_array.copy() if output_array is None else output_array

    def __init__(self, np.ndarray input_array, output_array=None, int axis=0,
                 int diagonals=8, domains=None, levels=None, int maxs=100, int use_direct=-1):
        cdef:
            int level
            list si = [slice(None)]*input_array.ndim

        FMMLevel.__init__(self, input_array.shape[axis], diagonals=diagonals,
                          domains=domains, levels=levels, maxs=maxs, use_direct=use_direct)
        self._a = _Lambda(np.arange(self.Nn, dtype='d'))
        self.axis = axis
        si[axis] = 0
        self.si = tuple(si)

    def __call__(self, np.ndarray input_array, output_array=None, int transpose=0):

        if input_array.shape[self.axis] <= self.use_direct and input_array.ndim == 1:
            self._output_array[...] = 0
            _leg2cheb(input_array, self._output_array, self._a, transpose)
            if output_array is not None:
                output_array[...] = self._output_array
                return output_array
            return self._output_array

        if transpose == 1:
            input_array = input_array.copy() # preserve input_array
            input_array *= (2/np.pi)
            input_array[self.si] *= 0.5

        if input_array.shape[self.axis] <= self.use_direct:
            self._output_array[...] = 0
            FMMdirect1(input_array, self._output_array, self.axis, self._a, self.N, transpose)
        else:
            self.apply(input_array, self._output_array, transpose)

        if transpose == 0:
            self._output_array *= (2/np.pi)
            self._output_array[self.si] *= 0.5

        if output_array is not None:
            output_array[...] = self._output_array
            return output_array
        return self._output_array

    cdef apply(self, np.ndarray input_array, np.ndarray output_array, int transpose):
        FMMcheb(input_array, output_array, self.axis, self.Nn, self.fk, self.Nk, self.T, self.Th, self.ThT, self.D, self.Mmin, self.s, self.diags, transpose)
        FMMdirect2(input_array, output_array, self.axis, self._a, 2*self.s, get_number_of_blocks(self.L, <long*>np.PyArray_DATA(self.D), len(self.D)), 2*self.diags, transpose)
        FMMdirect1(input_array, output_array, self.axis, self._a, self.diags, transpose)


cdef class Cheb2Leg(FMMLevel):
    cdef public np.ndarray _output_array
    cdef np.ndarray _a
    cdef np.ndarray _dn
    cdef int axis
    cdef tuple s0
    cdef tuple sn1

    def __cinit__(self, np.ndarray input_array, output_array=None, int axis=0,
                  int diagonals=8, domains=None, levels=None, int maxs=100, int use_direct=-1):
        self._output_array = input_array.copy() if output_array is None else output_array

    def __init__(self, np.ndarray input_array, output_array=None, int axis=0,
                 int diagonals=8, domains=None, levels=None, int maxs=100, int use_direct=-1):
        cdef:
            list si = [None]*input_array.ndim
            list sn1 = [slice(None)]*input_array.ndim
            np.ndarray[double] k

        FMMLevel.__init__(self, input_array.shape[axis], diagonals=diagonals,
                          domains=domains, levels=levels, l2c=0, maxs=maxs, use_direct=use_direct)

        k = np.arange(self.Nn, dtype='d')
        k[0] = 1
        self._dn = _Lambda((k[::2]-2)/2)/k[::2]
        self._a = 1/(2*_Lambda(k)*k*(k+0.5))
        self._a[0] = 2/np.sqrt(np.pi)
        self.axis = axis

        # precompute some multidimensional slices
        sn1[axis] = slice(1, self.N)
        self.sn1 = tuple(sn1)
        si[axis] = slice(None)
        self.s0 = tuple(si)

    def __call__(self, np.ndarray input_array, output_array=None):
        cdef:
            np.ndarray w0

        if input_array.shape[self.axis] <= self.use_direct and input_array.ndim == 1:
            self._output_array[...] = 0
            _cheb2leg(input_array, self._output_array, self._dn, self._a)
            if output_array is not None:
                output_array[...] = self._output_array
                return output_array
            return self._output_array
        w0 = input_array.copy()
        w0[self.sn1] *= np.arange(1, self.N)[self.s0]
        if input_array.shape[self.axis] <= self.use_direct:
            self._output_array[...] = 0
            FMMdirect4(w0, self._output_array, self.axis, self._dn[:self.N//2], self._a[:self.N], self.N//2)
        else:
            self.apply(w0, self._output_array)
        self._output_array *= (np.arange(self.N)+0.5)[self.s0]
        if output_array is not None:
            output_array[...] = self._output_array
            return output_array
        return self._output_array

    def apply(self, np.ndarray input_array, np.ndarray output_array):
        FMMcheb(input_array, output_array, self.axis, self.Nn, self.fk, self.Nk, self.T, self.Th, self.ThT, self.D, self.Mmin, self.s, self.diags, False)
        FMMdirect3(input_array, output_array, self.axis, self._dn, self._a, 2*self.s, get_number_of_blocks(self.L, <long*>np.PyArray_DATA(self.D), len(self.D)), 2*self.diags)
        FMMdirect4(input_array, output_array, self.axis, self._dn, self._a, self.diags)



import numpy as np
import numba as nb
from .diagma import *
from .threedma import *
from .twodma import *
from .tdma import *
from .pdma import *
from .fdma import *
from .heptadma import *
from .la import *
from .helmholtz import *
from .biharmonic import *
from .chebyshev import *
from .transforms import *

@nb.jit(nopython=True, fastmath=True, cache=True)
def crossND(c, a, b):
    c[0] = a[1]*b[2] - a[2]*b[1]
    c[1] = a[2]*b[0] - a[0]*b[2]
    c[2] = a[0]*b[1] - a[1]*b[0]
    return c

@nb.jit(nopython=True, fastmath=True, cache=True)
def cross2D(c, a, b):
    for i in range(a.shape[1]):
        for j in range(a.shape[2]):
                a0 = a[0, i, j]
                a1 = a[1, i, j]
                b0 = b[0, i, j]
                b1 = b[1, i, j]
                c[i, j] = a0*b1 - a1*b0

@nb.jit(nopython=True, fastmath=True, cache=True)
def cross3D(c, a, b):
    for i in range(a.shape[1]):
        for j in range(a.shape[2]):
            for k in range(a.shape[3]):
                a0 = a[0, i, j, k]
                a1 = a[1, i, j, k]
                a2 = a[2, i, j, k]
                b0 = b[0, i, j, k]
                b1 = b[1, i, j, k]
                b2 = b[2, i, j, k]
                c[0, i, j, k] = a1*b2 - a2*b1
                c[1, i, j, k] = a2*b0 - a0*b2
                c[2, i, j, k] = a0*b1 - a1*b0

@nb.jit(nopython=True, fastmath=True, cache=True)
def outer2D(a, b, c, symmetric):
    N, M = a.shape[1:]
    if symmetric:
        for i in range(N):
            for j in range(M):
                c[0, i, j] = a[0, i, j]**2           # (0, 0)
                c[1, i, j] = a[0, i, j]*a[1, i, j]   # (0, 1)
                c[2, i, j] = c[1, i, j]              # (1, 0)
                c[3, i, j] = a[1, i, j]**2           # (1, 1)
    else:
        for i in range(N):
            for j in range(M):
                c[0, i, j] = a[0, i, j]*b[0, i, j]   # (0, 0)
                c[1, i, j] = a[0, i, j]*b[1, i, j]   # (0, 1)
                c[2, i, j] = a[1, i, j]*b[0, i, j]   # (1, 0)
                c[3, i, j] = a[1, i, j]*b[1, i, j]   # (1, 1)


@nb.jit(nopython=True, fastmath=True, cache=True)
def outer3D(a, b, c, symmetric):
    N, M, P = a.shape[1:]
    if symmetric:
        for i in range(N):
            for j in range(M):
                for k in range(P):
                    c[0, i, j, k] = a[0, i, j, k]**2           # (0, 0)
                    c[1, i, j, k] = a[0, i, j, k]*a[1, i, j, k]   # (0, 1)
                    c[2, i, j, k] = a[0, i, j, k]*a[2, i, j, k]   # (0, 2)
                    c[3, i, j, k] = c[1, i, j, k]              # (1, 0)
                    c[4, i, j, k] = a[1, i, j, k]**2           # (1, 1)
                    c[5, i, j, k] = a[1, i, j, k]*a[2, i, j, k]   # (1, 2)
                    c[6, i, j, k] = c[2, i, j, k]              # (2, 0)
                    c[7, i, j, k] = c[5, i, j, k]              # (2, 1)
                    c[8, i, j, k] = a[2, i, j, k]**2           # (2, 2)
    else:
        for i in range(N):
            for j in range(M):
                for k in range(P):
                    c[0, i, j, k] = a[0, i, j, k]*b[0, i, j, k]   # (0, 0)
                    c[1, i, j, k] = a[0, i, j, k]*b[1, i, j, k]   # (0, 1)
                    c[2, i, j, k] = a[0, i, j, k]*b[2, i, j, k]   # (0, 2)
                    c[3, i, j, k] = a[1, i, j, k]*b[0, i, j, k]   # (1, 0)
                    c[4, i, j, k] = a[1, i, j, k]*b[1, i, j, k]   # (1, 1)
                    c[5, i, j, k] = a[1, i, j, k]*b[2, i, j, k]   # (1, 2)
                    c[6, i, j, k] = a[2, i, j, k]*b[0, i, j, k]   # (2, 0)
                    c[7, i, j, k] = a[2, i, j, k]*b[1, i, j, k]   # (2, 1)
                    c[8, i, j, k] = a[2, i, j, k]*b[2, i, j, k]   # (2, 2)


def apply_mask(u_hat, mask):
    if mask is not None:
        if u_hat.ndim == mask.ndim:
            mask = np.broadcast_to(mask, u_hat.shape)
            if mask.ndim == 1:
                u_hat = apply_mask_1D(u_hat, mask)
            elif mask.ndim == 2:
                u_hat = apply_mask_2D(u_hat, mask)
            elif mask.ndim == 3:
                u_hat = apply_mask_3D(u_hat, mask)
            elif mask.ndim == 4:
                u_hat = apply_mask_4D(u_hat, mask)
            else:
                u_hat *= mask
        elif u_hat.ndim == mask.ndim + 1:
            mask = np.broadcast_to(mask, u_hat.shape[1:])
            if mask.ndim == 1:
                u_hat = apply_bmask_1D(u_hat, mask)
            elif mask.ndim == 2:
                u_hat = apply_bmask_2D(u_hat, mask)
            elif mask.ndim == 3:
                u_hat = apply_bmask_3D(u_hat, mask)
            elif mask.ndim == 4:
                u_hat = apply_bmask_4D(u_hat, mask)
            else:
                u_hat *= mask
        else:
            u_hat = apply_bxmask(u_hat, mask)
    return u_hat

@nb.jit(nopython=True, fastmath=True, cache=True)
def apply_mask_1D(u, mask):
    for i in range(u.shape[0]):
        if mask[i] == 0:
            u[i] = 0
    return u

@nb.jit(nopython=True, fastmath=True, cache=True)
def apply_mask_2D(u, mask):
    for i in range(u.shape[0]):
        for j in range(u.shape[1]):
            if mask[i, j] == 0:
                u[i, j] = 0
    return u

@nb.jit(nopython=True, fastmath=True, cache=True)
def apply_mask_3D(u, mask):
    for i in range(u.shape[0]):
        for j in range(u.shape[1]):
            for k in range(u.shape[2]):
                if mask[i, j, k] == 0:
                    u[i, j, k] = 0
    return u

@nb.jit(nopython=True, fastmath=True, cache=True)
def apply_mask_4D(u, mask):
    for i in range(u.shape[0]):
        for j in range(u.shape[1]):
            for k in range(u.shape[2]):
                for l in range(u.shape[3]):
                    if mask[i, j, k, l] == 0:
                        u[i, j, k, l] = 0
    return u

@nb.jit(nopython=True, fastmath=True, cache=True)
def apply_bmask_1D(u, mask):
    for j in range(u.shape[1]):
        if mask[j] == 0:
            for i in range(u.shape[0]):
                u[i, j] = 0
    return u

@nb.jit(nopython=True, fastmath=True, cache=True)
def apply_bmask_2D(u, mask):
    for j in range(u.shape[1]):
        for k in range(u.shape[2]):
            if mask[j, k] == 0:
                for i in range(u.shape[0]):
                    u[i, j, k] = 0
    return u

@nb.jit(nopython=True, fastmath=True, cache=True)
def apply_bmask_3D(u, mask):
    for j in range(u.shape[1]):
        for k in range(u.shape[2]):
            for l in range(u.shape[3]):
                if mask[j, k, l] == 0:
                    for i in range(u.shape[0]):
                        u[i, j, k, l] = 0
    return u

@nb.jit(nopython=True, fastmath=True, cache=True)
def apply_bmask_4D(u, mask):
    for j in range(u.shape[1]):
        for k in range(u.shape[2]):
            for l in range(u.shape[3]):
                for m in range(u.shape[4]):
                    if mask[j, k, l, m] == 0:
                        for i in range(u.shape[0]):
                            u[i, j, k, l, m] = 0
    return u


@nb.jit(nopython=False, fastmath=True, cache=True)
def apply_bxmask(u_hat, mask):
    if mask is not None:
        N = mask.shape
        if len(N) == 1:
            mask = np.broadcast_to(mask, u_hat.shape[-1])
            for i in range(u_hat.shape[-1]):
                if mask[i] == 0:
                    u_hat[..., i] = 0
        elif len(N) == 2:
            mask = np.broadcast_to(mask, u_hat.shape[-2:])
            for i in range(u_hat.shape[-2]):
                for j in range(u_hat.shape[-1]):
                    if mask[i, j] == 0:
                        u_hat[..., i, j] = 0
        elif len(N) == 3:
            mask = np.broadcast_to(mask, u_hat.shape[-3:])
            for i in range(u_hat.shape[-3]):
                for j in range(u_hat.shape[-2]):
                    for k in range(u_hat.shape[-1]):
                        if mask[i, j, k] == 0:
                            u_hat[..., i, j, k] = 0
        elif len(N) == 4:
            mask = np.broadcast_to(mask, u_hat.shape[-4:])
            for i in range(u_hat.shape[-4]):
                for j in range(u_hat.shape[-3]):
                    for k in range(u_hat.shape[-2]):
                        for l in range(u_hat.shape[-1]):
                            if mask[i, j, k, l] == 0:
                                u_hat[..., i, j, k, l] = 0
        else:
            u_hat *= mask
    return u_hat



import numpy as np
import numba as nb

__all__ = ['SolverGeneric1ND_solve_data',
           'Solve_axis_2D', 'Solve_axis_3D', 'Solve_axis_4D']

@nb.njit
def SolverGeneric1ND_solve_data(u, data, sol, naxes, is_zero_index):
    if u.ndim == 2:
        if naxes == 0:
            for i in range(u.shape[1]):
                if i == 0 and is_zero_index:
                    continue
                sol(u[:, i], data[i])

        elif naxes == 1:
            for i in range(u.shape[0]):
                if i == 0 and is_zero_index:
                    continue
                sol(u[i], data[i])

    elif u.ndim == 3:
        if naxes == 0:
            for i in range(u.shape[1]):
                for j in range(u.shape[2]):
                    if i == 0 and j == 0 and is_zero_index:
                        continue
                    sol(u[:, i, j], data[i, j])

        elif naxes == 1:
            for i in range(u.shape[0]):
                for j in range(u.shape[2]):
                    if i == 0 and j == 0 and is_zero_index:
                        continue
                    sol(u[i, :, j], data[i, j])

        elif naxes == 2:
            for i in range(u.shape[0]):
                for j in range(u.shape[1]):
                    if i == 0 and j == 0 and is_zero_index:
                        continue
                    sol(u[i, j, :], data[i, j])
    return u

@nb.jit(nopython=True, fastmath=True, cache=False)
def Solve_axis_2D(data, x, innerfun, axis):
    if axis == 0:
        for j in range(x.shape[1]):
            innerfun(x[:, j], data)
    elif axis == 1:
        for i in range(x.shape[0]):
            innerfun(x[i, :], data)

@nb.jit(nopython=True, fastmath=True, cache=False)
def Solve_axis_3D(data, x, innerfun, axis):
    if axis == 0:
        for j in range(x.shape[1]):
            for k in range(x.shape[2]):
                innerfun(x[:, j, k], data)
    elif axis == 1:
        for i in range(x.shape[0]):
            for k in range(x.shape[2]):
                innerfun(x[i, :, k], data)
    elif axis == 2:
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                innerfun(x[i, j], data)

@nb.jit(nopython=True, fastmath=True, cache=False)
def Solve_axis_4D(data, x, innerfun, axis):
    if axis == 0:
        for j in range(x.shape[1]):
            for k in range(x.shape[2]):
                for l in range(x.shape[3]):
                    innerfun(x[:, j, k, l], data)
    elif axis == 1:
        for i in range(x.shape[0]):
            for k in range(x.shape[2]):
                for l in range(x.shape[3]):
                    innerfun(x[i, :, k, l], data)
    elif axis == 2:
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                for l in range(x.shape[3]):
                    innerfun(x[i, j, :, l], data)
    elif axis == 3:
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                for k in range(x.shape[2]):
                    innerfun(x[i, j, k], data)



r"""
Module for function spaces using Fourier exponentials.

A basis function :math:`\phi_k` is given as

.. math::
    \phi_k(x) = \exp(ikx)

and an expansion is given as

.. math::
    :label: u

    u(x) = \sum_{k=-N/2}^{N/2-1} \hat{u}_k \exp(ikx)

However, since :math:`\exp(ikx) = \exp(i(k \pm N)x)` this expansion can
also be written as an interpolator

.. math::
    :label: u2

    u(x) = \sum_{k=-N/2}^{N/2} \frac{\hat{u}_k}{c_k} \exp(ikx)

where :math:`c_{N/2} = c_{-N/2} = 2`, whereas :math:`c_k = 1` for
:math:`k=-N/2+1, ..., N/2-1`. Furthermore,
:math:`\hat{u}_{N/2} = \hat{u}_{-N/2}`.

The interpolator form is used for computing odd derivatives. Otherwise,
it makes no difference and therefore :eq:`u` is used in transforms, since
this is the form expected by fftw.
The inner product is defined as

.. math::
    (u, v) = \frac{1}{L} \int_{0}^{L} u \overline{v} dx

where :math:`\overline{v}` is the complex conjugate of :math:`v`, and
:math:`L` is the length of the (periodic) domain.
"""
import sympy as sp
import numpy as np
from mpi4py_fft import fftw
from shenfun.spectralbase import SpectralBase, Transform, islicedict, slicedict
from shenfun.optimization.cython import convolve
from shenfun.config import config

bases = ['R2C', 'C2C']
bcbases = []
testbases = []
__all__ = bases

#pylint: disable=method-hidden, no-member, line-too-long, arguments-differ


class FourierBase(SpectralBase):
    r"""Abstract base class for Fourier exponentials
    """
    def __init__(self, N, padding_factor=1, domain=(0, 2*sp.pi), dtype=float,
                 dealias_direct=False, coordinates=None):
        self._k = None
        self._planned_axes = None  # Collapsing of axes means that this base can be used to plan transforms over several collapsed axes. Store the axes planned for here.
        SpectralBase.__init__(self, N, dtype=dtype, padding_factor=padding_factor,
                              dealias_direct=dealias_direct, domain=domain,
                              coordinates=coordinates)

    @staticmethod
    def family():
        return 'fourier'

    @staticmethod
    def boundary_condition():
        return 'Periodic'

    def points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if N is None:
            N = self.shape(False)
        points = np.arange(N, dtype=float)*2*np.pi/N
        if map_true_domain is True:
            points = self.map_true_domain(points)
        if weighted:
            # weight is 1/self.domain_length() since this leads to the Kronecker delta function for mass matrix
            return points, np.array([float(self.domain_factor())/N])
        return points, np.array([2*np.pi/N])

    def orthogonal_basis_function(self, i=0, x=sp.symbols('x', real=True)):
        k = self.wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
        return sp.exp(sp.I*k[i]*x)

    def L2_norm_sq(self, i):
        return 1

    def l2_norm_sq(self, i=None):
        if i is None:
            return np.ones(self.N)
        return 1

    def weight(self, x=sp.symbols('x', real=True)):
        return 1/self.domain_length()

    def evaluate_basis(self, x=None, i=0, output_array=None):
        if x is None:
            x = self.mesh(False, False)
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape, dtype=complex)

        if self._k is None:
            self._k = self.wavenumbers(bcast=False)
        k = self._k[i]
        output_array[:] = np.exp(1j*x*k)
        return output_array

    def evaluate_basis_derivative(self, x=None, i=0, k=0, output_array=None):
        output_array = self.evaluate_basis(x, i, output_array)
        l = self._k[i]
        output_array *= ((1j*l)**k)
        return output_array

    def vandermonde(self, x):
        k = self.wavenumbers(bcast=False)
        x = np.atleast_1d(x)
        return np.exp(1j*x[:, np.newaxis]*k[np.newaxis, :])

    def evaluate_basis_derivative_all(self, x=None, k=0):
        V = self.evaluate_basis_all(x=x)
        if k > 0:
            l = self.wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
            V = V*((1j*l)**k)[np.newaxis, :]
        return V

    # Reimplemented for efficiency (smaller array in *= when truncated)
    def forward(self, input_array=None, output_array=None, kind='fast'):
        if kind != 'fast':
            return SpectralBase.forward(self, input_array, output_array, kind=kind)

        if input_array is not None:
            self.forward.input_array[...] = input_array

        self.forward.xfftn()
        self._truncation_forward(self.forward.tmp_array,
                                 self.forward.output_array)
        M = self.get_normalization()
        self.forward._output_array *= M

        self.apply_inverse_mass(self.forward.output_array)

        if output_array is not None:
            output_array[...] = self.forward.output_array
            return output_array
        return self.forward.output_array

    def apply_inverse_mass(self, array):
        coors = self.tensorproductspace.coors if self.tensorproductspace else self.coors
        if not coors.is_cartesian: # mass matrix may not be diagonal, or there is scaling
            return SpectralBase.apply_inverse_mass(self, array)
        return array

    def _evaluate_scalar_product(self, kind='fast'):
        if kind != 'fast':
            SpectralBase._evaluate_scalar_product(self, kind=kind)
            return
        output = self.scalar_product.xfftn()
        output *= self.get_normalization()

    def reference_domain(self):
        return (0, 2*sp.pi)

    @property
    def is_orthogonal(self):
        return True

    def get_orthogonal(self, **kwargs):
        return self.get_unplanned(**kwargs)

    def get_homogeneous(self, **kwargs):
        return self.get_unplanned(**kwargs)

    def get_unplanned(self, **kwargs):
        d = dict(domain=self.domain,
                 padding_factor=self.padding_factor,
                 dealias_direct=self.dealias_direct,
                 coordinates=self.coors.coordinates)
        d.update(kwargs)
        return self.__class__(self.N, **d)

    def get_dealiased(self, padding_factor=1.5, dealias_direct=False, **kwargs):
        d = dict(domain=self.domain,
                 padding_factor=padding_factor,
                 dealias_direct=dealias_direct,
                 coordinates=self.coors.coordinates)
        d.update(kwargs)
        return self.__class__(self.N, **d)

    def get_refined(self, N, **kwargs):
        d = dict(domain=self.domain,
                 padding_factor=self.padding_factor,
                 dealias_direct=self.dealias_direct,
                 coordinates=self.coors.coordinates)
        d.update(kwargs)
        return self.__class__(N, **d)

    def mask_nyquist(self, u_hat, mask=None):
        """Return array `u_hat` with zero Nyquist coefficients

        Parameters
        ----------
        u_hat : array
            Array to be masked
        mask : array or None, optional
            mask array, if not provided then get the mask by calling
            :func:`get_mask_nyquist`
        """
        if mask is None:
            mask = self.get_mask_nyquist(bcast=False)
        if mask is not None:
            u_hat *= mask
        return u_hat

    def plan(self, shape, axis, dtype, options):
        if shape in (0, (0,)):
            return

        if isinstance(axis, int):
            axis = [axis]
        s = list(np.take(shape, axis))

        if isinstance(self.forward, Transform):
            if self.forward.input_array.shape == shape and axis == self._planned_axes:
                # Already planned
                return

        plan_fwd = self._xfftn_fwd
        plan_bck = self._xfftn_bck

        self.axis = axis[-1]
        self._planned_axes = axis

        #opts = dict(
        #    overwrite_input='FFTW_DESTROY_INPUT',
        #    planner_effort=self.opts['planner_effort'],
        #    threads=self.opts['threads'],
        #)
        opts = plan_fwd.opts
        opts['overwrite_input'] = 'FFTW_DESTROY_INPUT'
        opts.update(options)
        threads = opts['threads']
        flags = (fftw.flag_dict[opts['planner_effort']],
                 fftw.flag_dict[opts['overwrite_input']])

        U = fftw.aligned(shape, dtype=dtype)
        xfftn_fwd = plan_fwd(U, s=s, axes=axis, threads=threads, flags=flags)
        V = xfftn_fwd.output_array

        opts = plan_bck.opts
        opts['overwrite_input'] = 'FFTW_DESTROY_INPUT'
        opts.update(options)
        threads = opts['threads']
        flags = (fftw.flag_dict[opts['planner_effort']],
                 fftw.flag_dict[opts['overwrite_input']])
        if np.issubdtype(dtype, np.floating):
            flags = (fftw.flag_dict[opts['planner_effort']],)

        xfftn_bck = plan_bck(V, s=s, axes=axis, threads=threads, flags=flags, output_array=U)
        V.fill(0)
        U.fill(0)
        self._M = xfftn_fwd.get_normalization()

        if self.padding_factor > 1.+1e-8:
            trunc_array = self._get_truncarray(shape, V.dtype)
            self.scalar_product = Transform(self.scalar_product, xfftn_fwd, U, V, trunc_array)
            self.forward = Transform(self.forward, xfftn_fwd, U, V, trunc_array)
            self.backward = Transform(self.backward, xfftn_bck, trunc_array, V, U)

        else:
            self.scalar_product = Transform(self.scalar_product, xfftn_fwd, U, V, V)
            self.forward = Transform(self.forward, xfftn_fwd, U, V, V)
            self.backward = Transform(self.backward, xfftn_bck, V, V, U)

        self.si = islicedict(axis=self.axis, dimensions=self.dimensions)
        self.sl = slicedict(axis=self.axis, dimensions=self.dimensions)


class R2C(FourierBase):
    r"""Fourier function space for real to complex transforms

    A basis function :math:`\phi_k` is given as

    .. math::
        \phi_k(x) = \exp(ikx), \quad k =-N/2, -N/2+1, \ldots, N/2-1

    An expansion is given as

    .. math::
        u(x) = \sum_{k=-N/2}^{N/2-1} \hat{u}_k \exp(ikx).

    Howewer, due to Hermitian symmetry :math:`\hat{u}_k = \overline{\hat{u}_{-k}}`,
    which is taken advantage of in this function space. Specifically, Fourier
    transforms make use of real-to-complex and complex-to-real algorithms, see
    `FFTW <https://www.fftw.org/fftw3_doc/One_002dDimensional-DFTs-of-Real-Data.html>`_
    and `rfftn <https://mpi4py-fft.readthedocs.io/en/latest/mpi4py_fft.fftw.html#mpi4py_fft.fftw.xfftn.rfftn>`_
    of `mpi4py-fft <https://github.com/mpi4py/mpi4py-fft>`_.

    Parameters
    ----------
    N : int
        Number of quadrature points. Should be even for efficiency, but
        this is not required.
    padding_factor : float, optional
        Factor for padding backward transforms. padding_factor=1.5
        corresponds to a 3/2-rule for dealiasing.
    domain : 2-tuple of numbers, optional
        The computational domain.
    dealias_direct : bool, optional
        True for dealiasing using 2/3-rule. Must be used with
        padding_factor = 1.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """

    def __init__(self, N, padding_factor=1., domain=(0, 2*sp.pi),
                 dealias_direct=False, coordinates=None, **kw):
        FourierBase.__init__(self, N, padding_factor=padding_factor, dtype=float,
                             domain=domain, dealias_direct=dealias_direct,
                             coordinates=coordinates)
        self.N = N
        self._xfftn_fwd = fftw.rfftn
        self._xfftn_bck = fftw.irfftn
        self._xfftn_fwd.opts = config['fftw']['rfft']
        self._xfftn_bck.opts = config['fftw']['irfft']
        self._sn = []
        self._sm = []
        self.plan((int(padding_factor*N),), (0,), float, {})

    def wavenumbers(self, bcast=True, scaled=False, eliminate_highest_freq=False):
        k = np.fft.rfftfreq(self.N, 1./self.N).astype(int)
        if self.N % 2 == 0 and eliminate_highest_freq:
            k[-1] = 0
        if scaled:
            k = k*float(self.domain_factor())
        if bcast is True:
            k = self.broadcast_to_ndims(k)
        return k

    def get_mask_nyquist(self, bcast=True):
        """Return None or an array with zeros for Nyquist coefficients and one otherwise

        Parameters
        ----------
        bcast : boolean, optional
            If True then broadcast returned mask array to dimensions of the
            :class:`TensorProductSpace` this base belongs to.
        """
        if self.N % 2 == 0:
            f = np.ones(self.N//2+1, dtype=int)
            f[-1] = 0
        else:
            return None
        if bcast is True:
            f = self.broadcast_to_ndims(f)
        return f

    def _get_truncarray(self, shape, dtype):
        shape = list(shape)
        shape[self.axis] = int(shape[self.axis] / self.padding_factor)
        shape[self.axis] = shape[self.axis]//2 + 1
        return fftw.aligned(shape, dtype=dtype)

    @staticmethod
    def short_name():
        return 'R2C'

    def slice(self):
        return slice(0, self.N//2+1)

    def shape(self, forward_output=True):
        if forward_output:
            return self.N//2+1
        return int(np.floor(self.padding_factor*self.N))

    def _evaluate_expansion_all(self, input_array, output_array, x=None, kind='fast'):
        if kind == 'vandermonde':
            assert abs(self.padding_factor-1) < 1e-8
            P = self.evaluate_basis_all(x=x)
            if output_array.ndim == 1:
                output_array[:] = np.dot(P, input_array).real
                if self.N % 2 == 0:
                    output_array += np.conj(np.dot(P[:, 1:-1], input_array[1:-1])).real
                else:
                    output_array += np.conj(np.dot(P[:, 1:], input_array[1:])).real

            else:
                fc = np.moveaxis(input_array, self.axis, -2)
                array = np.dot(P, fc).real
                s = [slice(None)]*fc.ndim
                if self.N % 2 == 0:
                    s[-2] = slice(1, -1)
                    array += np.conj(np.dot(P[:, 1:-1], fc[tuple(s)])).real
                else:
                    s[-2] = slice(1, None)
                    array += np.conj(np.dot(P[:, 1:], fc[tuple(s)])).real

                output_array[:] = np.moveaxis(array, 0, self.axis)
            return
        assert kind == 'fast'
        assert input_array is self.backward.tmp_array
        assert output_array is self.backward.output_array
        self.backward.xfftn(normalise_idft=False)

    def _truncation_forward(self, padded_array, trunc_array):
        if not id(trunc_array) == id(padded_array):
            trunc_array.fill(0)
            N = trunc_array.shape[self.axis]
            s = self.sl[slice(0, N)]
            trunc_array[:] = padded_array[s]
            if self.N % 2 == 0:
                s1 = self.si[N-1]
                trunc_array[s1] = trunc_array[s1].real
                trunc_array[s1] *= 2

    def _padding_backward(self, trunc_array, padded_array):
        if not id(trunc_array) == id(padded_array):
            padded_array.fill(0)
            N = trunc_array.shape[self.axis]
            if len(self._sn) != self.dimensions:
                self._sn = self.sl[slice(0, N)]
                self._sm = self.si[N-1]
            padded_array[self._sn] = trunc_array[self._sn]
            if self.N % 2 == 0:  # Symmetric Fourier interpolator
                padded_array[self._sm] = padded_array[self._sm].real
                padded_array[self._sm] *= 0.5

        elif self.dealias_direct:
            N = self.N
            su = self.sl[slice(N//3, None)]
            padded_array[su] = 0

    def convolve(self, u, v, uv=None, fast=True):
        """Convolution of u and v.

        Parameters
        ----------
        u : array
        v : array
        uv : array, optional
        fast : bool, optional
               Whether to use fast transforms in computing convolution

        Note
        ----
        Note that this method is only valid for 1D data, and that
        for multidimensional arrays one should use corresponding method
        in the :class:`.TensorProductSpace` class.

        """
        N = self.N

        if fast is True:
            if uv is None:
                uv = self.forward.output_array.copy()

            assert self.padding_factor > 1.0, "padding factor must be > 3/2+1/N to perform convolution without aliasing"
            u2 = self.backward.output_array.copy()
            u3 = self.backward.output_array.copy()
            u2 = self.backward(u, u2)
            u3 = self.backward(v, u3)
            uv = self.forward(u2*u3, uv)

        else:
            if uv is None:
                uv = np.zeros(N+1, dtype=u.dtype)
            Np = N if not N % 2 == 0 else N+1
            k1 = np.fft.fftfreq(Np, 1./Np).astype(int)
            convolve.convolve_real_1D(u, v, uv, k1)

        return uv


class C2C(FourierBase):
    r"""Fourier function space for complex to complex transforms

    A basis function :math:`\phi_k` is given as

    .. math::
        \phi_k(x) = \exp(ikx), \quad k =-N/2, -N/2+1, \ldots, N/2-1

    An expansion is given as

    .. math::
        u(x) = \sum_{k=-N/2}^{N/2-1} \hat{u}_k \exp(ikx).

    Parameters
    ----------
    N : int
        Number of quadrature points. Should be even for efficiency, but
        this is not required.
    padding_factor : float, optional
        Factor for padding backward transforms. padding_factor=1.5
        corresponds to a 3/2-rule for dealiasing.
    domain : 2-tuple of numbers, optional
        The computational domain.
    dealias_direct : bool, optional
        True for dealiasing using 2/3-rule. Must be used with
        padding_factor = 1.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """

    def __init__(self, N, padding_factor=1, domain=(0, 2*sp.pi),
                 dealias_direct=False, coordinates=None, **kw):
        FourierBase.__init__(self, N, padding_factor=padding_factor, dtype=complex,
                             domain=domain, dealias_direct=dealias_direct,
                             coordinates=coordinates)
        self.N = N
        self._xfftn_fwd = fftw.fftn
        self._xfftn_bck = fftw.ifftn
        self._xfftn_fwd.opts = config['fftw']['fft']
        self._xfftn_bck.opts = config['fftw']['ifft']
        self.plan((int(padding_factor*N),), (0,), complex, {})
        self._slp = []

    @staticmethod
    def short_name():
        return 'C2C'

    def _evaluate_expansion_all(self, input_array, output_array, x=None, kind='fast'):
        if kind == 'vandermonde':
            SpectralBase._evaluate_expansion_all(self, input_array, output_array, x, kind=kind)
            return
        assert kind == 'fast'
        assert input_array is self.backward.tmp_array
        assert output_array is self.backward.output_array
        self.backward.xfftn(normalise_idft=False)

    def wavenumbers(self, bcast=True, scaled=False, eliminate_highest_freq=False):
        k = np.fft.fftfreq(self.N, 1./self.N).astype(int)
        if self.N % 2 == 0 and eliminate_highest_freq:
            k[self.N//2] = 0
        if scaled:
            k = k*float(self.domain_factor())
        if bcast is True:
            k = self.broadcast_to_ndims(k)
        return k

    def get_mask_nyquist(self, bcast=True):
        if self.N % 2 == 0:
            f = np.ones(self.N, dtype=int)
            f[self.N//2] = 0
        else:
            return None
        if bcast is True:
            f = self.broadcast_to_ndims(f)
        return f

    def slice(self):
        return slice(0, self.N)

    def shape(self, forward_output=True):
        if forward_output:
            return self.N
        return int(np.floor(self.padding_factor*self.N))

    def count_trailing_zeros(self, u, reltol=1e-12, abstol=1e-15):
        assert u.function_space() == self
        assert u.ndim == 1
        a = abs(u[self.slice()])
        ua = (a < reltol*a.max()) | (a < abstol)
        return np.argmin(ua[self.N//2:]) + np.argmin(ua[:self.N//2][::-1])

    def _truncation_forward(self, padded_array, trunc_array):
        if not id(trunc_array) == id(padded_array):
            trunc_array.fill(0)
            N = trunc_array.shape[self.axis]
            su = self.sl[slice(0, N//2+1)]
            trunc_array[su] = padded_array[su]
            su = self.sl[slice(-(N//2), None)]
            trunc_array[su] += padded_array[su]

    def _padding_backward(self, trunc_array, padded_array):
        # pylint: disable=attribute-defined-outside-init
        if not id(trunc_array) == id(padded_array):
            padded_array.fill(0)
            N = trunc_array.shape[self.axis]
            if len(self._slp) != self.dimensions: # Store for microoptimization
                self._slp = self.sl[slice(0, N//2+1)]
                self._slm = self.sl[slice(-(N//2), None)]
                self._slp0 = self.si[N//2]
                self._slm0 = self.si[-(N//2)]
            padded_array[self._slp] = trunc_array[self._slp]
            padded_array[self._slm] = trunc_array[self._slm]
            if self.N % 2 == 0:  # Use symmetric Fourier interpolator
                padded_array[self._slp0] *= 0.5
                padded_array[self._slm0] *= 0.5

        elif self.dealias_direct:
            N = trunc_array.shape[self.axis]
            su = self.sl[slice(N//3, -(N//3)+1)]
            padded_array[su] = 0

    def convolve(self, u, v, uv=None, fast=True):
        """Convolution of u and v.

        Parameters
        ----------
            u : array
            v : array
            uv : array, optional
            fast : bool, optional
                   Whether to use fast transforms in computing convolution

        Note
        ----
        Note that this method is only valid for 1D data, and that
        for multidimensional arrays one should use corresponding method
        in the TensorProductSpace class.

        """
        assert len(u.shape) == 1
        N = self.N

        if fast:
            if uv is None:
                uv = self.forward.output_array.copy()

            assert self.padding_factor > 1.0, "padding factor must be > 3/2+1/N to perform convolution without aliasing"
            u2 = self.backward.output_array.copy()
            u3 = self.backward.output_array.copy()
            u2 = self.backward(u, u2)
            u3 = self.backward(v, u3)
            uv = self.forward(u2*u3, uv)

        else:

            if uv is None:
                uv = np.zeros(2*N, dtype=u.dtype)

            Np = N if not N % 2 == 0 else N+1
            k = np.fft.fftfreq(Np, 1./Np).astype(int)
            convolve.convolve_1D(u, v, uv, k)

        return uv



from __future__ import print_function
from time import time
import functools
from itertools import product
import pytest
import numpy as np
from mpi4py import MPI
from sympy import symbols, cos, sin, lambdify, exp
from shenfun.config import config
from shenfun.fourier import bases as fbases
from shenfun.chebyshev import bases as cbases
from shenfun.chebyshevu import bases as cubases
from shenfun.ultraspherical import bases as ubases
from shenfun.legendre import bases as lbases
from shenfun.laguerre import bases as lagbases
from shenfun.hermite import bases as hbases
from shenfun.jacobi import bases as jbases
from shenfun import Function, project, Dx, Array, FunctionSpace, TensorProductSpace, \
   VectorSpace, CompositeSpace, inner

comm = MPI.COMM_WORLD

for f in ['dct', 'dst', 'fft', 'ifft', 'rfft', 'irfft']:
    config['fftw'][f]['planner_effort'] = 'FFTW_ESTIMATE'

abstol = dict(f=5e-3, d=1e-10, g=1e-12)

def allclose(a, b):
    atol = abstol[a.dtype.char.lower()]
    return np.allclose(a, b, rtol=0, atol=atol)

def random_like(array):
    shape = array.shape
    dtype = array.dtype
    return np.random.random(shape).astype(dtype)

sizes = (12, 13)
@pytest.mark.parametrize('typecode', 'fFdD')
@pytest.mark.parametrize('dim', (2, 3, 4))
def test_transform(typecode, dim):
    s = (True,)
    if comm.Get_size() > 2 and dim > 2:
        s = (True, False)

    for slab in s:
        for shape in product(*([sizes]*dim)):
            bases = []
            for n in shape[:-1]:
                bases.append(FunctionSpace(n, 'F', dtype=typecode.upper()))
            bases.append(FunctionSpace(shape[-1], 'F', dtype=typecode))

            fft = TensorProductSpace(comm, bases, dtype=typecode, slab=slab)

            if comm.rank == 0:
                grid = [c.size for c in fft.subcomm]
                print('grid:{} shape:{} typecode:{}'
                      .format(grid, shape, typecode))

            U = random_like(fft.forward.input_array)

            F = fft.forward(U)
            V = fft.backward(F)
            assert allclose(V, U)

            # Alternative method
            fft.forward.input_array[...] = U
            fft.forward(kind={'fourier': 'vandermonde'})
            fft.backward(kind={'fourier': 'vandermonde'})
            V = fft.backward.output_array
            assert allclose(V, U)

            TT = VectorSpace(fft)
            U = Array(TT)
            V = Array(TT)
            F = Function(TT)
            U[:] = random_like(U)
            F = TT.forward(U, F)
            V = TT.backward(F, V)
            assert allclose(V, U)

            TM = CompositeSpace([fft, fft])
            U = Array(TM)
            V = Array(TM)
            F = Function(TM)
            U[:] = random_like(U)
            F = TM.forward(U, F)
            V = TM.backward(F, V)
            assert allclose(V, U)

            fftp = fft.get_dealiased(padding_factor=1.5)

            #fft.destroy()

            #padding = 1.5
            #bases = []
            #for n in shape[:-1]:
            #    bases.append(FunctionSpace(n, 'F', dtype=typecode.upper(), padding_factor=padding))
            #bases.append(FunctionSpace(shape[-1], 'F', dtype=typecode, padding_factor=padding))

            #fft = TensorProductSpace(comm, bases, dtype=typecode)

            if comm.rank == 0:
                grid = [c.size for c in fftp.subcomm]
                print('grid:{} shape:{} typecode:{}'
                      .format(grid, shape, typecode))

            U = random_like(fftp.forward.input_array)
            F = fftp.forward(U)

            Fc = F.copy()
            V = fftp.backward(F)
            F = fftp.forward(V)
            assert allclose(F, Fc)

            # Alternative method
            fftp.backward.input_array[...] = F
            fftp.backward()
            fftp.forward()
            V = fftp.forward.output_array
            assert allclose(F, V)

            fftp.destroy()
            fft.destroy()

ctrialBasis = [cbases.__dict__.get(base) for base in cbases.bases[:-1]]
ctestBasis = ctrialBasis + [cbases.__dict__.get(base) for base in cbases.testbases]
cutrialBasis = [cubases.__dict__.get(base) for base in cubases.bases[:-1]]
cutestBasis = cutrialBasis + [cubases.__dict__.get(base) for base in cubases.testbases]

utrialBasis = [ubases.__dict__.get(base) for base in ubases.bases[:-1]]
utestBasis = utrialBasis + [ubases.__dict__.get(base) for base in ubases.testbases]
ltrialBasis = [lbases.__dict__.get(base) for base in lbases.bases[:-1]]
ltestBasis = ltrialBasis + [lbases.__dict__.get(base) for base in lbases.testbases]
latrialBasis = [lagbases.__dict__.get(base) for base in lagbases.bases[:-1]]
htrialBasis = [hbases.__dict__.get(base) for base in hbases.bases]
jtrialBasis = [jbases.__dict__.get(base) for base in jbases.bases[:-1]]
jtestBasis = jtrialBasis + [jbases.__dict__.get(base) for base in jbases.testbases]


cquads = ('GC', 'GL')
lquads = ('LG', 'GL')
lagquads = ('LG',)
hquads = ('HG',)
jquads = ('JG',)
uquads = ('QG',)
cuquads = ('GU', 'GC')

all_bases_and_quads = (list(product(cutrialBasis, cuquads))
                       +list(product(utrialBasis, uquads))
                       +list(product(ltrialBasis, lquads))
                       +list(product(ctrialBasis, cquads))
                       +list(product(jtrialBasis, jquads))
                       )
lag_bases_and_quads = list(product(latrialBasis, lagquads))
h_bases_and_quads = list(product(htrialBasis, hquads))

num_tests = 20
some_bases_and_quads = [all_bases_and_quads[i] for i in np.random.randint(0, len(all_bases_and_quads), num_tests)]

@pytest.mark.parametrize('typecode', 'dD')
@pytest.mark.parametrize('dim', (1, 2))
@pytest.mark.parametrize('ST,quad', some_bases_and_quads
                                    +lag_bases_and_quads
                                    +h_bases_and_quads)
def test_shentransform(typecode, dim, ST, quad):
    sizes = (16, 15)
    for shape in product(*([sizes]*dim)):
        bases = []
        for n in shape[:-1]:
            bases.append(FunctionSpace(n, 'F', dtype=typecode.upper()))
        bases.append(FunctionSpace(shape[-1], 'F', dtype=typecode))

        for axis in range(dim+1):
            ST0 = ST(sizes[-1]+4, quad=quad)
            bases.insert(axis, ST0)
            fft = TensorProductSpace(comm, bases, dtype=typecode)
            U = random_like(fft.forward.input_array)
            F = fft.forward(U)
            Fc = F.copy()
            V = fft.backward(F)
            F = fft.forward(U)
            assert allclose(F, Fc)
            bases.pop(axis)
            fft.destroy()

bases_and_quads = (list(product(ltrialBasis[:2], lquads))
                   +list(product(ctrialBasis[:2], cquads)))
                   #+list(product(jBasis[:2], jquads)))

axes = {2: {0: [0, 1, 2],
            1: [1, 0, 2],
            2: [2, 0, 1]},
        1: {0: [0, 1],
            1: [1, 0]}}

@pytest.mark.parametrize('typecode', 'dD')
@pytest.mark.parametrize('dim', (1, 2))
@pytest.mark.parametrize('ST,quad', bases_and_quads)
def test_project(typecode, dim, ST, quad):
    # Using sympy to compute an analytical solution
    x, y, z = symbols("x,y,z")
    sizes = (10, 9)

    funcs = {
        (1, 0): (cos(1*y)*sin(1*np.pi*x))*(1-x**2),
        (1, 1): (cos(1*x)*sin(1*np.pi*y))*(1-y**2),
        (2, 0): (sin(1*z)*cos(1*y)*sin(1*np.pi*x))*(1-x**2),
        (2, 1): (sin(1*z)*cos(1*x)*sin(1*np.pi*y))*(1-y**2),
        (2, 2): (sin(1*x)*cos(1*y)*sin(1*np.pi*z))*(1-z**2)
        }
    xs = {0:x, 1:y, 2:z}

    for shape in product(*([sizes]*dim)):
        bases = []
        for n in shape[:-1]:
            bases.append(FunctionSpace(n, 'F', dtype=typecode.upper()))
        bases.append(FunctionSpace(shape[-1], 'F', dtype=typecode))

        for axis in range(dim+1):
            ST0 = ST(2*shape[-1], quad=quad)
            bases.insert(axis, ST0)
            fft = TensorProductSpace(comm, bases, dtype=typecode, axes=axes[dim][axis])
            dfft = fft.get_orthogonal()
            X = fft.local_mesh(True)
            ue = funcs[(dim, axis)]
            uq = Array(fft, buffer=ue)
            uh = Function(fft)
            uh = fft.forward(uq, uh)
            due = ue.diff(xs[axis], 1)
            duq = Array(fft, buffer=due)
            dh = project(Dx(uh, axis, 1), dfft)
            uf = dh.backward()
            assert np.linalg.norm(uf-duq) < 1e-3, np.linalg.norm(uf-duq)
            for ax in (x for x in range(dim+1) if x is not axis):
                due = ue.diff(xs[axis], 1, xs[ax], 1)
                duq = Array(fft, buffer=due)
                uf = project(Dx(Dx(uh, axis, 1), ax, 1), dfft).backward()
                assert np.linalg.norm(uf-duq) < 1e-3

            bases.pop(axis)
            fft.destroy()
            dfft.destroy()

#lagbases_and_quads = list(product(lagBasis, lagquads))
@pytest.mark.parametrize('typecode', 'dD')
@pytest.mark.parametrize('dim', (1, 2))
#@pytest.mark.parametrize('ST,quad', lagbases_and_quads)
def test_project_lag(typecode, dim):
    # Using sympy to compute an analytical solution
    x, y, z = symbols("x,y,z")
    sizes = (20, 17)

    funcs = {
        (1, 0): (cos(4*y)*sin(2*x))*exp(-x),
        (1, 1): (cos(4*x)*sin(2*y))*exp(-y),
        (2, 0): (sin(3*z)*cos(4*y)*sin(2*x))*exp(-x),
        (2, 1): (sin(2*z)*cos(4*x)*sin(2*y))*exp(-y),
        (2, 2): (sin(2*x)*cos(4*y)*sin(2*z))*exp(-z)
        }
    xs = {0:x, 1:y, 2:z}

    for shape in product(*([sizes]*dim)):
        bases = []
        for n in shape[:-1]:
            bases.append(FunctionSpace(n, 'F', dtype=typecode.upper()))
        bases.append(FunctionSpace(shape[-1], 'F', dtype=typecode))

        for axis in range(dim+1):
            ST1 = latrialBasis[1](3*shape[-1])
            bases.insert(axis, ST1)
            fft = TensorProductSpace(comm, bases, dtype=typecode, axes=axes[dim][axis])
            dfft = fft.get_orthogonal()
            X = fft.local_mesh(True)
            ue = funcs[(dim, axis)]
            due = ue.diff(xs[0], 1)
            u_h = project(ue, fft)
            du_h = project(due, dfft)
            du2 = project(Dx(u_h, 0, 1), dfft)
            uf = u_h.backward()
            assert np.linalg.norm(du2-du_h) < 1e-3
            bases.pop(axis)
            fft.destroy()
            dfft.destroy()

#hbases_and_quads = list(product(hBasis, hquads))
@pytest.mark.parametrize('typecode', 'dD')
@pytest.mark.parametrize('dim', (1, 2))
#@pytest.mark.parametrize('ST,quad', hbases_and_quads)
def test_project_hermite(typecode, dim):
    # Using sympy to compute an analytical solution
    x, y, z = symbols("x,y,z")
    sizes = (10, 9)

    funcs = {
        (1, 0): (cos(4*y)*sin(2*x))*exp(-x**2/2),
        (1, 1): (cos(4*x)*sin(2*y))*exp(-y**2/2),
        (2, 0): (sin(3*z)*cos(4*y)*sin(2*x))*exp(-x**2/2),
        (2, 1): (sin(2*z)*cos(4*x)*sin(2*y))*exp(-y**2/2),
        (2, 2): (sin(2*x)*cos(4*y)*sin(2*z))*exp(-z**2/2)
        }
    xs = {0:x, 1:y, 2:z}

    for shape in product(*([sizes]*dim)):
        bases = []
        for n in shape[:-1]:
            bases.append(FunctionSpace(n, 'F', dtype=typecode.upper()))
        bases.append(FunctionSpace(shape[-1], 'F', dtype=typecode))

        for axis in range(dim+1):
            ST0 = htrialBasis[0](3*shape[-1])
            bases.insert(axis, ST0)
            fft = TensorProductSpace(comm, bases, dtype=typecode, axes=axes[dim][axis])
            X = fft.local_mesh(True)
            ue = funcs[(dim, axis)]
            due = ue.diff(xs[0], 1)
            u_h = project(ue, fft)
            du_h = project(due, fft)
            du2 = project(Dx(u_h, 0, 1), fft)

            assert np.linalg.norm(du_h-du2) < 1e-5
            bases.pop(axis)
            fft.destroy()

@pytest.mark.parametrize('quad', lquads)
def test_project_2dirichlet(quad):
    x, y = symbols("x,y")
    ue = (cos(4*y)*sin(2*x))*(1-x**2)*(1-y**2)
    sizes = (18, 17)

    D0 = lbases.ShenDirichlet(sizes[0], quad=quad)
    D1 = lbases.ShenDirichlet(sizes[1], quad=quad)
    B0 = lbases.Orthogonal(sizes[0], quad=quad)
    B1 = lbases.Orthogonal(sizes[1], quad=quad)

    DD = TensorProductSpace(comm, (D0, D1))
    BD = TensorProductSpace(comm, (B0, D1))
    DB = TensorProductSpace(comm, (D0, B1))
    BB = TensorProductSpace(comm, (B0, B1))

    X = DD.local_mesh(True)
    uh = Function(DD, buffer=ue)
    dudx_hat = project(Dx(uh, 0, 1), BD)
    dx = Function(BD, buffer=ue.diff(x, 1))
    assert np.allclose(dx, dudx_hat, 0, 1e-5)

    dudy = project(Dx(uh, 1, 1), DB).backward()
    duedy = Array(DB, buffer=ue.diff(y, 1))
    assert np.allclose(duedy, dudy, 0, 1e-5)

    us_hat = Function(BB)
    uq = uh.backward()
    us = project(uq, BB, output_array=us_hat).backward()
    assert np.allclose(us, uq, 0, 1e-5)
    dudxy = project(Dx(us_hat, 0, 1) + Dx(us_hat, 1, 1), BB).backward()
    dxy = Array(BB, buffer=ue.diff(x, 1) + ue.diff(y, 1))
    assert np.allclose(dxy, dudxy, 0, 1e-5), np.linalg.norm(dxy-dudxy)
    DD.destroy()
    BD.destroy()
    DB.destroy()
    BB.destroy()

@pytest.mark.parametrize('typecode', 'dD')
@pytest.mark.parametrize('dim', (1, 2))
@pytest.mark.parametrize('ST,quad', some_bases_and_quads)
def test_eval_tensor(typecode, dim, ST, quad):
    # Using sympy to compute an analytical solution
    # Testing for Dirichlet and regular basis
    x, y, z = symbols("x,y,z")
    sizes = (28, 27)

    funcx = (1-x**2)**3*sin(np.pi*x)
    funcy = (1-y**2)**3*sin(np.pi*y)
    funcz = (1-z**2)**3*sin(np.pi*z)

    funcs = {
        (1, 0): cos(y)*funcx,
        (1, 1): cos(x)*funcy,
        (2, 0): sin(z)*cos(y)*funcx,
        (2, 1): sin(z)*cos(x)*funcy,
        (2, 2): sin(x)*cos(y)*funcz
        }
    syms = {1: (x, y), 2:(x, y, z)}
    points = None
    if comm.Get_rank() == 0:
        points = np.random.random((dim+1, 4))
    points = comm.bcast(points)
    t_0 = 0
    t_1 = 0
    t_2 = 0
    for shape in product(*([sizes]*dim)):
        #for shape in ((64, 64),):
        bases = []
        for n in shape[:-1]:
            bases.append(FunctionSpace(n, 'F', dtype=typecode.upper()))
        bases.append(FunctionSpace(shape[-1], 'F', dtype=typecode))

        for axis in range(dim+1):
            #for axis in (0,):
            ST0 = ST(shape[-1]+6, quad=quad)
            bases.insert(axis, ST0)
            # Spectral space must be aligned in nonperiodic direction, hence axes
            fft = TensorProductSpace(comm, bases, dtype=typecode, axes=axes[dim][axis])
            #print('axes', axes[dim][axis])
            #print('bases', bases)
            #print(bases[0].axis, bases[1].axis)
            X = fft.local_mesh(True)
            ue = funcs[(dim, axis)]
            ul = lambdify(syms[dim], ue, 'numpy')
            uq = ul(*points).astype(typecode)
            u_hat = Function(fft, buffer=ue)
            t0 = time()
            result = fft.eval(points, u_hat, method=0)
            t_0 += time()-t0
            assert np.allclose(uq, result, 0, 1e-2), np.linalg.norm(uq-result)
            t0 = time()
            result = fft.eval(points, u_hat, method=1)
            t_1 += time()-t0
            assert np.allclose(uq, result, 0, 1e-2)
            t0 = time()
            result = fft.eval(points, u_hat, method=2)
            t_2 += time()-t0
            assert np.allclose(uq, result, 0, 1e-2), uq/result
            result = u_hat.eval(points)
            assert np.allclose(uq, result, 0, 1e-2)

            bases.pop(axis)
            fft.destroy()
    print('method=0', t_0)
    print('method=1', t_1)
    print('method=2', t_2)

@pytest.mark.parametrize('typecode', 'dD')
@pytest.mark.parametrize('dim', (2, 3))
def test_eval_fourier(typecode, dim):
    # Using sympy to compute an analytical solution
    x, y, z = symbols("x,y,z")
    sizes = (12, 11)

    funcs = {
        2: cos(4*x) + sin(3*y),
        3: sin(3*x) + cos(4*y) + sin(2*z)
        }
    syms = {2: (x, y), 3: (x, y, z)}
    points = None
    if comm.Get_rank() == 0:
        points = np.random.random((dim, 3))
    points = comm.bcast(points)
    t_0 = 0
    t_1 = 0
    t_2 = 0
    for shape in product(*([sizes]*dim)):
        bases = []
        for n in shape[:-1]:
            bases.append(FunctionSpace(n, 'F', dtype=typecode.upper()))

        bases.append(FunctionSpace(shape[-1], 'F', dtype=typecode))
        fft = TensorProductSpace(comm, bases, dtype=typecode)
        X = fft.local_mesh(True)
        ue = funcs[dim]
        ul = lambdify(syms[dim], ue, 'numpy')
        uq = ul(*points).astype(typecode)
        u_hat = Function(fft, buffer=ue)
        t0 = time()
        result = fft.eval(points, u_hat, method=0)
        t_0 += time() - t0
        assert allclose(uq, result), print(uq, result)
        t0 = time()
        result = fft.eval(points, u_hat, method=1)
        t_1 += time() - t0
        assert allclose(uq, result)
        t0 = time()
        result = fft.eval(points, u_hat, method=2)
        t_2 += time() - t0
        assert allclose(uq, result), print(uq, result)
        fft.destroy()

    print('method=0', t_0)
    print('method=1', t_1)
    print('method=2', t_2)

@pytest.mark.parametrize('f0,f1', product(*([('C', 'L', 'F')])*2))
def test_inner(f0, f1):
    if f0 == 'F' and f1 == 'F':
        B0 = FunctionSpace(8, f0, dtype='D', domain=(-2*np.pi, 2*np.pi))
    else:
        B0 = FunctionSpace(8, f0)
    c = Array(B0, val=1)
    d = inner(1, c)
    assert abs(d-(B0.domain[1]-B0.domain[0])) < 1e-7

    B1 = FunctionSpace(8, f1)
    T = TensorProductSpace(comm, (B0, B1))
    a0 = Array(T, val=1)
    c0 = inner(1, a0)
    L = np.array([b.domain[1]-b.domain[0] for b in (B0, B1)])
    assert abs(c0-np.prod(L)) < 1e-7
    T.destroy()

    if not (f0 == 'F' or f1 == 'F'):
        B2 = FunctionSpace(8, f1, domain=(-2, 2))
        T = TensorProductSpace(comm, (B0, B1, B2))
        a0 = Array(T, val=1)
        c0 = inner(1, a0)
        L = np.array([b.domain[1]-b.domain[0] for b in (B0, B1, B2)])
        assert abs(c0-np.prod(L)) < 1e-7
        T.destroy()

@pytest.mark.parametrize('fam', ('C', 'L', 'F', 'La', 'H'))
def test_assign(fam):
    x, y = symbols("x,y")
    for bc in (None, (0, 0), (0, 0, 0, 0)):
        dtype = 'D' if fam == 'F' else 'd'
        bc = 'periodic' if fam == 'F' else bc
        if bc in ((0, 0), (0, 0, 0, 0)) and fam in ('La', 'H'):
            continue
        tol = 1e-12 if fam in ('C', 'L', 'F') else 1e-5
        N = (10, 12)
        B0 = FunctionSpace(N[0], fam, dtype=dtype, bc=bc)
        B1 = FunctionSpace(N[1], fam, dtype=dtype, bc=bc)
        u_hat = Function(B0)
        u_hat[1:4] = 1
        ub_hat = Function(B1)
        u_hat.assign(ub_hat)
        assert abs(inner(1, u_hat)-inner(1, ub_hat)) < tol
        T = TensorProductSpace(comm, (B0, B1))
        u_hat = Function(T)
        u_hat[1:4, 1:4] = 1
        Tp = T.get_refined((2*N[0], 2*N[1]))
        ub_hat = Function(Tp)
        u_hat.assign(ub_hat)
        assert abs(inner(1, u_hat)-inner(1, ub_hat)) < tol
        VT = VectorSpace(T)
        u_hat = Function(VT)
        u_hat[:, 1:4, 1:4] = 1
        Tp = T.get_refined((2*N[0], 2*N[1]))
        VTp = VectorSpace(Tp)
        ub_hat = Function(VTp)
        u_hat.assign(ub_hat)
        assert abs(inner((1, 1), u_hat)-inner((1, 1), ub_hat)) < tol
        T.destroy()
        Tp.destroy()

def test_refine():
    assert comm.Get_size() < 7
    N = (8, 9, 10)
    F0 = FunctionSpace(8, 'F', dtype='D')
    F1 = FunctionSpace(9, 'F', dtype='D')
    F2 = FunctionSpace(10, 'F', dtype='d')
    T = TensorProductSpace(comm, (F0, F1, F2), slab=True, collapse_fourier=True)
    u_hat = Function(T)
    u = Array(T)
    u[:] = np.random.random(u.shape)
    u_hat = u.forward(u_hat)
    Tp = T.get_dealiased(padding_factor=(2, 2, 2))
    u_ = Array(Tp)
    up_hat = Function(Tp)
    assert up_hat.commsizes == u_hat.commsizes
    u2 = u_hat.refine(2*np.array(N))
    V = VectorSpace(T)
    u_hat = Function(V)
    u = Array(V)
    u[:] = np.random.random(u.shape)
    u_hat = u.forward(u_hat)
    Vp = V.get_dealiased(padding_factor=(2, 2, 2))
    u_ = Array(Vp)
    up_hat = Function(Vp)
    assert up_hat.commsizes == u_hat.commsizes
    u3 = u_hat.refine(2*np.array(N))
    T.destroy()
    Tp.destroy()

def test_eval_expression():
    import sympy as sp
    from shenfun import div, grad
    x, y, z = sp.symbols('x,y,z')
    B0 = FunctionSpace(16, 'C')
    B1 = FunctionSpace(17, 'C')
    B2 = FunctionSpace(20, 'F', dtype='d')

    TB = TensorProductSpace(comm, (B0, B1, B2))
    f = sp.sin(x)+sp.sin(y)+sp.sin(z)
    dfx = f.diff(x, 2) + f.diff(y, 2) + f.diff(z, 2)
    fa = Function(TB, buffer=f)

    dfe = div(grad(fa))
    dfa = project(dfe, TB)

    xyz = np.array([[0.25, 0.5, 0.75],
                    [0.25, 0.5, 0.75],
                    [0.25, 0.5, 0.75]])

    f0 = lambdify((x, y, z), dfx)(*xyz)
    f1 = dfe.eval(xyz)
    f2 = dfa.eval(xyz)
    assert np.allclose(f0, f1, 1e-7)
    assert np.allclose(f1, f2, 1e-7)
    TB.destroy()

@pytest.mark.parametrize('fam', ('C', 'L'))
def test_eval_expression2(fam):
    import sympy as sp
    from shenfun import div, grad
    x, y = sp.symbols('x,y')
    B0 = FunctionSpace(16, fam, domain=(-2, 2))
    B1 = FunctionSpace(17, fam, domain=(-1, 3))

    T = TensorProductSpace(comm, (B0, B1))
    f = sp.sin(x)+sp.sin(y)
    dfx = f.diff(x, 2) + f.diff(y, 2)
    fa = Function(T, buffer=f)

    dfe = div(grad(fa))
    dfa = project(dfe, T)

    xy = np.array([[0.25, 0.5, 0.75],
                   [0.25, 0.5, 0.75]])

    f0 = lambdify((x, y), dfx)(*xy)
    f1 = dfe.eval(xy)
    f2 = dfa.eval(xy)
    assert np.allclose(f0, f1, 1e-7)
    assert np.allclose(f1, f2, 1e-7)
    T.destroy()


if __name__ == '__main__':
    test_transform('F', 2)
    #test_transform('d', 2)
    #test_shentransform('d', 2, jbases.Phi4, 'JG')
    #test_eval_expression()
    #test_eval_expression2('L')
    test_project('d', 2, ltrialBasis[0], 'GL')
    #test_project_lag('d', 2)
    #test_project_hermite('d', 2)
    #test_project2('d', 2, lbases.ShenNeumann, 'LG')
    #test_project_2dirichlet('GL')
    #test_eval_tensor('d', 2, cbases.ShenDirichlet, 'GC')
    #test_eval_fourier('D', 3)
    #test_inner('C', 'F')
    #test_refine()
    #test_assign('C')
